from __future__ import absolute_import, division, print_function

from ..accumulators import Mean, WeightedMean, WeightedSum

import numpy as np


class View(np.ndarray):
    __slots__ = ()

    def __getitem__(self, ind):
        sliced = super(View, self).__getitem__(ind)

        # If the shape is empty, return the parent type
        if not sliced.shape:
            return self._PARENT._make(*sliced)
        # If the dtype has changed, return a normal array (no longer a record)
        elif sliced.dtype != self.dtype:
            return np.asarray(sliced)
        # Otherwise, no change, return the same View type
        else:
            return sliced

    def __repr__(self):
        # Numpy starts the ndarray class name with "array", so we replace it
        # with our class name
        return (
            "{self.__class__.__name__}(\n      ".format(self=self)
            + repr(self.view(np.ndarray))[6:]
        )

    def __str__(self):
        fields = ", ".join(self._FIELDS)
        return "{self.__class__.__name__}: ({fields})\n{arr}".format(
            self=self, fields=fields, arr=self.view(np.ndarray)
        )


def make_getitem_property(name):
    def fget(self):
        return self[name]

    def fset(self, value):
        self[name] = value

    return property(fget, fset)


def fields(*names):
    """
    This decorator adds the name to the _FIELDS
    class property (for printing in reprs), and
    adds a property that looks like this:

    @property
    def name(self):
        return self["name"]
    @name.setter
    def name(self, value):
        self["name"] = value
    """

    def injector(cls):
        if hasattr(cls, "_FIELDS"):
            raise RuntimeError(
                "{0} already has had a fields decorator applied".format(
                    self.__class__.__name__
                )
            )
        fields = []
        for name in names:
            fields.append(name)
            setattr(cls, name, make_getitem_property(name))
            cls._FIELDS = tuple(fields)

        return cls

    return injector


@fields("value", "variance")
class WeightedSumView(View):
    __slots__ = ()
    _PARENT = WeightedSum


@fields(
    "sum_of_weights",
    "sum_of_weights_squared",
    "value",
    "sum_of_weighted_deltas_squared",
)
class WeightedMeanView(View):
    __slots__ = ()
    _PARENT = WeightedMean

    @property
    def variance(self):
        return self["sum_of_weighted_deltas_squared"] / (
            self["sum_of_weights"]
            - self["sum_of_weights_squared"] / self["sum_of_weights"]
        )


@fields("count", "value", "sum_of_deltas_squared")
class MeanView(View):
    __slots__ = ()
    _PARENT = Mean

    # Variance is a computation
    @property
    def variance(self):
        return self["sum_of_deltas_squared"] / (self["count"] - 1)


def _to_view(item, value=False):
    for cls in View.__subclasses__():
        if cls._FIELDS == item.dtype.names:
            ret = item.view(cls)
            if value and ret.shape:
                return ret.value
            else:
                return ret
    return item



from pytest import approx
import boost_histogram as bh


def test_mean_hist():

    h = bh.Histogram(bh.axis.Regular(3, 0, 1), storage=bh.storage.Mean())

    h.fill(0.10, sample=[2.5])
    h.fill(0.25, sample=[3.5])
    h.fill(0.45, sample=[1.2])
    h.fill(0.51, sample=[3.4])
    h.fill(0.81, sample=[1.3])
    h.fill(0.86, sample=[1.9])

    results = (
        dict(count=2, value=3.0, variance=0.5),
        dict(count=2, value=2.3, variance=2.42),
        dict(count=2, value=1.6, variance=0.18),
    )

    for i in range(len(h.axes[0])):

        assert results[i]["count"] == h[i].count
        assert results[i]["value"] == approx(h[i].value)
        assert results[i]["variance"] == approx(h[i].variance)



from __future__ import absolute_import, division, print_function

del absolute_import, division, print_function

from .._core.axis import options

from .._core import axis as ca

from .kwargs import KWArgs
from .sig_tools import inject_signature
from .axis_transform import AxisTransform
from .utils import cast, register, set_family, MAIN_FAMILY, CPP_FAMILY, set_module

import copy


# Contains common methods and properties to all axes
@set_module("boost_histogram.axis")
class Axis(object):
    __slots__ = ("_ax",)

    def __copy__(self):
        other = self.__class__.__new__(self.__class__)
        other._ax = copy.copy(self._ax)
        return other

    def index(self, value):
        """
        Return the fractional index(es) given a value (or values) on the axis.
        """

        return self._ax.index(value)

    def value(self, index):
        """
        Return the value(s) given an (fractional) index (or indices).
        """

        return self._ax.value(index)

    def bin(self, index):
        """
        Return the edges of the bins as a tuple for a
        continuous axis or the bin value for a
        non-continuous axis, when given an index.
        """

        return self._ax.bin(index)

    def __eq__(self, other):
        return self._ax == other._ax

    def __ne__(self, other):
        return self._ax != other._ax

    @property
    def metadata(self):
        """
        Get or set the metadata associated with this axis.
        """
        return self._ax.metadata

    @metadata.setter
    def metadata(self, value):
        self._ax.metadata = value

    @classmethod
    def _convert_cpp(cls, cpp_object):
        nice_ax = cls.__new__(cls)
        nice_ax._ax = cpp_object
        return nice_ax

    def __len__(self):
        return self._ax.size

    def __iter__(self):
        return self._ax.__iter__()


# Mixin for main style classes
# Contains common methods and properties to all Main module axes
class MainAxisMixin(object):
    __slots__ = ()

    def __repr__(self):
        return "{self.__class__.__name__}({args}{kwargs})".format(
            self=self, args=self._repr_args(), kwargs=self._repr_kwargs()
        )

    def _repr_kwargs(self):
        """
        Return options for use in repr. Metadata is last,
        just in case it spans multiple lines.
        """

        ret = ""
        if self.options.growth:
            ret += ", growth=True"
        elif self.options.circular:
            ret += ", circular=True"
        else:
            if not self.options.underflow:
                ret += ", underflow=False"
            if not self.options.overflow:
                ret += ", overflow=False"

        if self.metadata is not None:
            ret += ", metadata={0!r}".format(self.metadata)

        return ret

    @property
    def options(self):
        """
        Return the options.  Fields:
          .underflow - True if axes captures values that are too small
          .overflow  - True if axes captures values that are too large
                       (or non-valid for category axes)
          .growth    - True if axis can grow
          .circular  - True if axis wraps around
        """
        return self._ax.options

    @property
    def size(self):
        """
        Return number of bins excluding under- and overflow.
        """
        return self._ax.size

    @property
    def extent(self):
        """
        Return number of bins including under- and overflow.
        """
        return self._ax.extent

    def __getitem__(self, i):
        """
        Access a bin, using normal Python syntax for wraparound.
        """
        # UHI support
        if callable(i):
            i = i(self)
        else:
            if i < 0:
                i += self._ax.size
            if i >= self._ax.size:
                raise IndexError(
                    "Out of range access, {0} is more than {1}".format(i, self._ax.size)
                )
        return self.bin(i)

    @property
    def edges(self):
        return self._ax.edges

    @property
    def centers(self):
        """
        An array of bin centers.
        """
        return self._ax.centers

    @property
    def widths(self):
        """
        An array of bin widths.
        """
        return self._ax.widths


# Contains all common methods for cpp module axes
class CppAxisMixin(object):
    __slots__ = ()

    def __repr__(self):
        return repr(self._ax)

    def options(self):
        """
        Return the options.  Fields:
          .underflow - True if axes captures values that are too small
          .overflow  - True if axes captures values that are too large
                       (or non-valid for category axes)
          .growth    - True if axis can grow
          .circular  - True if axis wraps around
        """
        return self._ax.options

    def size(self):
        """
        Return number of bins excluding under- and overflow.
        """
        return self._ax.size

    def extent(self):
        """
        Return number of bins including under- and overflow.
        """
        return self._ax.extent

    def edges(self):
        return self._ax.edges

    def centers(self):
        """
        An array of bin centers.
        """
        return self._ax.centers

    def widths(self):
        """
        An array of bin widths.
        """
        return self._ax.widths


# Contains all common methods and properties for Regular axes
@register(
    {
        ca.regular_uoflow,
        ca.regular_uoflow_growth,
        ca.regular_uflow,
        ca.regular_oflow,
        ca.regular_none,
        ca.regular_numpy,
        ca.regular_pow,
        ca.regular_trans,
        ca.regular_circular,
    }
)
class BaseRegular(Axis):
    __slots__ = ()

    @inject_signature(
        "self, bins, start, stop, *, metadata=None, underflow=True, overflow=True, growth=False, circular=False, transform=None"
    )
    def __init__(self, bins, start, stop, **kwargs):
        """
        Make a regular axis with nice keyword arguments for underflow,
        overflow, and growth.

        Parameters
        ----------
        bins : int
            The number of bins between start and stop
        start : float
            The beginning value for the axis
        stop : float
            The ending value for the axis
        metadata : Any
            Any Python object to attach to the axis, like a label.
        underflow : bool = True
            Enable the underflow bin
        overflow : bool = True
            Enable the overflow bin
        growth : bool = False
            Allow the axis to grow if a value is encountered out of range.
            Be careful, the axis will grow as large as needed.
        circular : bool = False
            Filling wraps around.
        transform : Optional[AxisTransform] = None
            Transform the regular bins (Log, Sqrt, and Pow(v))
        """

        with KWArgs(kwargs) as k:
            metadata = k.optional("metadata")
            transform = k.optional("transform")
            options = k.options(
                underflow=True, overflow=True, growth=False, circular=False
            )

        if transform is not None:
            if options != {"underflow", "overflow"}:
                raise KeyError("Transform supplied, cannot change other options")

            if (
                not isinstance(transform, AxisTransform)
                and AxisTransform in transform.__bases__
            ):
                raise TypeError("You must pass an instance, use {}()".format(transform))

            self._ax = transform._produce(bins, start, stop, metadata)

        elif options == {"growth", "underflow", "overflow"}:
            self._ax = ca.regular_uoflow_growth(bins, start, stop, metadata)
        elif options == {"underflow", "overflow"}:
            self._ax = ca.regular_uoflow(bins, start, stop, metadata)
        elif options == {"underflow"}:
            self._ax = ca.regular_uflow(bins, start, stop, metadata)
        elif options == {"overflow"}:
            self._ax = ca.regular_oflow(bins, start, stop, metadata)
        elif options == {
            "circular",
            "underflow",
            "overflow",
        } or options == {  # growth=True should work
            "circular",
            "overflow",
        }:  # growth=True, underflow=False is also correct
            self._ax = ca.regular_circular(bins, start, stop, metadata)

        elif options == set():
            self._ax = ca.regular_none(bins, start, stop, metadata)
        else:
            raise KeyError("Unsupported collection of options")


@set_module("boost_histogram.axis")
@set_family(MAIN_FAMILY)
class Regular(BaseRegular, MainAxisMixin):
    __slots__ = ()

    def _repr_args(self):
        "Return inner part of signature for use in repr"

        return "{bins:g}, {start:g}, {stop:g}".format(
            bins=self.size, start=self.edges[0], stop=self.edges[-1]
        )

    def _repr_kwargs(self):
        ret = super(Regular, self)._repr_kwargs()

        if self.transform is not None:
            ret += ", transform={0}".format(self.transform)

        return ret

    @property
    def transform(self):
        if hasattr(self._ax, "transform"):
            return cast(self, self._ax.transform, AxisTransform)
        return None


@set_module("boost_histogram.cpp.axis")
@set_family(CPP_FAMILY)
class regular(BaseRegular, CppAxisMixin):
    __slots__ = ()

    def transform(self):
        if hasattr(self._ax, "transform"):
            return cast(self, self._ax.transform, AxisTransform)
        return None


@register(
    {
        ca.variable_none,
        ca.variable_uflow,
        ca.variable_oflow,
        ca.variable_uoflow,
        ca.variable_uoflow_growth,
        ca.variable_circular,
    }
)
class BaseVariable(Axis):
    __slots__ = ()

    @inject_signature(
        "self, edges, *, metadata=None, underflow=True, overflow=True, growth=False"
    )
    def __init__(self, edges, **kwargs):
        """
        Make an axis with irregularly spaced bins. Provide a list
        or array of bin edges, and len(edges)-1 bins will be made.

        Parameters
        ----------
        edges : Array[float]
            The edges for the bins. There will be one less bin than edges.
        metadata : object
            Any Python object to attach to the axis, like a label.
        underflow : bool = True
            Enable the underflow bin
        overflow : bool = True
            Enable the overflow bin
        circular : bool = False
            Enable wraparound
        growth : bool = False
            Allow the axis to grow if a value is encountered out of range.
            Be careful, the axis will grow as large as needed.
        """
        with KWArgs(kwargs) as k:
            metadata = k.optional("metadata")
            options = k.options(
                underflow=True, overflow=True, circular=False, growth=False
            )

        if options == {"growth", "underflow", "overflow"}:
            self._ax = ca.variable_uoflow_growth(edges, metadata)
        elif options == {"underflow", "overflow"}:
            self._ax = ca.variable_uoflow(edges, metadata)
        elif options == {"underflow"}:
            self._ax = ca.variable_uflow(edges, metadata)
        elif options == {"overflow"}:
            self._ax = ca.variable_oflow(edges, metadata)
        elif options == {
            "circular",
            "underflow",
            "overflow",
        } or options == {  # growth=True should work
            "circular",
            "overflow",
        }:  # growth=True, underflow=False is also correct
            self._ax = ca.variable_circular(edges, metadata)
        elif options == set():
            self._ax = ca.variable_none(edges, metadata)
        else:
            raise KeyError("Unsupported collection of options")


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.axis")
class Variable(BaseVariable, MainAxisMixin):
    __slots__ = ()

    def _repr_args(self):
        "Return inner part of signature for use in repr"

        if len(self) > 20:
            return repr(self.edges)
        else:
            return "[{}]".format(", ".join(format(v, "g") for v in self.edges))


@set_family(CPP_FAMILY)
@set_module("boost_histogram.cpp.axis")
class variable(BaseVariable, CppAxisMixin):
    __slots__ = ()


@register(
    {
        ca.integer_none,
        ca.integer_uflow,
        ca.integer_oflow,
        ca.integer_uoflow,
        ca.integer_growth,
        ca.integer_circular,
    }
)
class BaseInteger(Axis):
    __slots__ = ()

    @inject_signature(
        "self, start, stop, *, metadata=None, underflow=True, overflow=True, growth=False"
    )
    def __init__(self, start, stop, **kwargs):
        """
        Make an integer axis, with a collection of consecutive integers.

        Parameters
        ----------
        start : int
            The beginning value for the axis
        stop : int
            The ending value for the axis. (start-stop) bins will be created.
        metadata : object
            Any Python object to attach to the axis, like a label.
        underflow : bool = True
            Enable the underflow bin
        overflow : bool = True
            Enable the overflow bin
        circular : bool = False
            Enable wraparound
        growth : bool = False
            Allow the axis to grow if a value is encountered out of range.
            Be careful, the axis will grow as large as needed.
        """
        with KWArgs(kwargs) as k:
            metadata = k.optional("metadata")
            options = k.options(
                underflow=True, overflow=True, circular=False, growth=False
            )

        # underflow and overflow settings are ignored, integers are always
        # finite and thus cannot end up in a flow bin when growth is on
        if "growth" in options and "circular" not in options:
            self._ax = ca.integer_growth(start, stop, metadata)
        elif options == {"underflow", "overflow"}:
            self._ax = ca.integer_uoflow(start, stop, metadata)
        elif options == {"underflow"}:
            self._ax = ca.integer_uflow(start, stop, metadata)
        elif options == {"overflow"}:
            self._ax = ca.integer_oflow(start, stop, metadata)
        elif (
            "circular" in options and "growth" not in options
        ):  # growth=True should work
            self._ax = ca.integer_circular(
                start, stop, metadata
            )  # flow bins do no matter
        elif options == set():
            self._ax = ca.integer_none(start, stop, metadata)
        else:
            raise KeyError("Unsupported collection of options")


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.axis")
class Integer(BaseInteger, MainAxisMixin):
    __slots__ = ()

    def _repr_args(self):
        "Return inner part of signature for use in repr"

        return "{start:g}, {stop:g}".format(start=self.edges[0], stop=self.edges[-1])


@set_family(CPP_FAMILY)
@set_module("boost_histogram.cpp.axis")
class integer(BaseInteger, CppAxisMixin):
    __slots__ = ()


@register({ca.category_str_growth, ca.category_str})
class BaseStrCategory(Axis):
    __slots__ = ()

    @inject_signature("self, categories, *, metadata=None, growth=False")
    def __init__(self, categories, **kwargs):
        """
        Make a category axis with strings; items will
        be added to a predefined list of bins or a growing (with growth=True)
        list of bins.


        Parameters
        ----------
        categories : Iterator[str]
            The bin values in strings. May be empty if growth is enabled.
        metadata : object
            Any Python object to attach to the axis, like a label.
        growth : bool = False
            Allow the axis to grow if a value is encountered out of range.
            Be careful, the axis will grow as large as needed.
        """
        with KWArgs(kwargs) as k:
            metadata = k.optional("metadata")
            options = k.options(growth=False)

        # We need to make sure we support Python 2 for now :(
        # henryiii: This shortcut possibly should be removed
        if isinstance(categories, (type(""), type(u""))):
            categories = list(categories)

        if options == {"growth"}:
            self._ax = ca.category_str_growth(categories, metadata)
        elif options == set():
            self._ax = ca.category_str(categories, metadata)
        else:
            raise KeyError("Unsupported collection of options")


@register({ca.category_int, ca.category_int_growth})
class BaseIntCategory(Axis):
    __slots__ = ()

    @inject_signature("self, categories, *, metadata=None, growth=False")
    def __init__(self, categories, **kwargs):
        """
        Make a category axis with ints; items will
        be added to a predefined list of bins or a growing (with growth=True)
        list of bins. An empty list is allowed if growth=True.


        Parameters
        ----------
        categories : Iteratable[int]
            The bin values, either ints or strings.
        metadata : object
            Any Python object to attach to the axis, like a label.
        growth : bool = False
            Allow the axis to grow if a value is encountered out of range.
            Be careful, the axis will grow as large as needed.
        """
        with KWArgs(kwargs) as k:
            metadata = k.optional("metadata")
            options = k.options(growth=False)

        if options == {"growth"}:
            self._ax = ca.category_int_growth(categories, metadata)
        elif options == set():
            self._ax = ca.category_int(categories, metadata)
        else:
            raise KeyError("Unsupported collection of options")


class CategoryMixin(object):
    __slots__ = ()

    def _repr_kwargs(self):
        """
        Return options for use in repr. Metadata is last,
        just in case it spans multiple lines.


        This is specialized for Category axes to avoid repeating
        the flow arguments unnecessarily.
        """

        ret = ""
        if self.options.growth:
            ret += ", growth=True"
        elif self.options.circular:
            ret += ", circular=True"

        if self.metadata is not None:
            ret += ", metadata={0!r}".format(self.metadata)

        return ret


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.axis")
class StrCategory(BaseStrCategory, CategoryMixin, MainAxisMixin):
    __slots__ = ()

    def _repr_args(self):
        "Return inner part of signature for use in repr"

        return "[{0}]".format(", ".join(repr(c) for c in self))


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.axis")
class IntCategory(BaseIntCategory, CategoryMixin, MainAxisMixin):
    __slots__ = ()

    def _repr_args(self):
        "Return inner part of signature for use in repr"

        return "[{0}]".format(", ".join(format(c, "g") for c in self))


@set_family(CPP_FAMILY)
@set_module("boost_histogram.cpp.axis")
class int_category(BaseIntCategory, CppAxisMixin):
    __slots__ = ()


@set_family(CPP_FAMILY)
@set_module("boost_histogram.cpp.axis")
class str_category(BaseStrCategory, CppAxisMixin):
    __slots__ = ()



# To regenerate API docs:
# sphinx-apidoc -o api ../src/boost_histogram -M -f -t template/

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import re

import sys

DIR = os.path.abspath(os.path.dirname(__file__))
BASEDIR = os.path.abspath(os.path.dirname(DIR))
sys.path.append(os.path.join(BASEDIR, "src"))

# -- Project information -----------------------------------------------------

project = "boost_histogram"
copyright = "2019, Henry Schreiner, Hans Dembinski"
author = "Henry Schreiner, Hans Dembinski"

# It is better to use pkg_resources, but we can't build on RtD
from pkg_resources import get_distribution, DistributionNotFound

try:
    version = get_distribution("boost_histogram").version
except DistributionNotFound:
    pass  # No version (latest/git hash)


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "recommonmark",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "nbsphinx",
    "sphinx.ext.mathjax",
    "sphinx_copybutton",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "**.ipynb_checkpoints",
    "Thumbs.db",
    ".DS_Store",
    ".env",
    "api/boost_histogram.*.rst",
]

# Read the Docs needs this explicitly listed.
master_doc = "index"

# -- Options for Notebook input ----------------------------------------------

nbsphinx_execute = "never"  # Can change to auto

highlight_language = "python3"

nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'png2x'}",
    "--InlineBackend.rc={'figure.dpi': 96}",
]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []  # _static

# Simpler docs (no build required)

autodoc_mock_imports = ["boost_histogram._core"]



from __future__ import absolute_import, division, print_function

del absolute_import, division, print_function

__all__ = ("histogram", "axis", "storage", "accumulators", "algorithm", "__version__")

from .._internal.hist import histogram

from . import axis, storage, accumulators, algorithm
from .. import __version__



from __future__ import absolute_import, division, print_function

del absolute_import, division, print_function

from .._core import storage as store
from .utils import set_family, MAIN_FAMILY, CPP_FAMILY, set_module

# Simple mixin to provide a common base class for types
class Storage(object):
    def __repr__(self):
        return "{self.__class__.__name__}()".format(self=self)


### MAIN FAMILY


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.storage")
class Int64(store.int64, Storage):
    pass


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.storage")
class Double(store.double, Storage):
    pass


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.storage")
class AtomicInt64(store.atomic_int64, Storage):
    pass


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.storage")
class Unlimited(store.unlimited, Storage):
    pass


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.storage")
class Weight(store.weight, Storage):
    pass


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.storage")
class Mean(store.mean, Storage):
    pass


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.storage")
class WeightedMean(store.weighted_mean, Storage):
    pass


### CPP FAMILY ###


@set_family(CPP_FAMILY)
@set_module("boost_histogram.cpp.storage")
class int64(Int64):
    pass


@set_family(CPP_FAMILY)
@set_module("boost_histogram.cpp.storage")
class double(Double):
    pass


@set_family(CPP_FAMILY)
@set_module("boost_histogram.cpp.storage")
class atomic_int64(AtomicInt64):
    pass


@set_family(CPP_FAMILY)
@set_module("boost_histogram.cpp.storage")
class unlimited(Unlimited):
    pass


@set_family(CPP_FAMILY)
@set_module("boost_histogram.cpp.storage")
class weight(Weight):
    pass


@set_family(CPP_FAMILY)
@set_module("boost_histogram.cpp.storage")
class mean(Mean):
    pass


@set_family(CPP_FAMILY)
@set_module("boost_histogram.cpp.storage")
class weighted_mean(WeightedMean):
    pass



import pytest
import boost_histogram as bh
import numpy as np
from numpy.testing import assert_array_equal


@pytest.mark.parametrize(
    "storage",
    [bh.storage.Int64, bh.storage.Double, bh.storage.AtomicInt64, bh.storage.Unlimited],
)
def test_setting(storage):
    h = bh.Histogram(bh.axis.Regular(10, 0, 1), storage=storage())

    h[0] = 2
    h[1] = 3
    h[-1] = 5

    assert h[0] == 2
    assert h[1] == 3
    assert h[9] == 5

    assert_array_equal(h.view(), [2, 3, 0, 0, 0, 0, 0, 0, 0, 5])


def test_setting_weight():
    h = bh.Histogram(bh.axis.Regular(10, 0, 10), storage=bh.storage.Weight())

    h.fill([0.3, 0.3, 0.4, 1.2])

    assert h[0] == bh.accumulators.WeightedSum(3, 3)
    assert h[1] == bh.accumulators.WeightedSum(1, 1)

    h[0] = bh.accumulators.WeightedSum(value=2, variance=2)
    assert h[0] == bh.accumulators.WeightedSum(2, 2)

    a = h.view()

    assert a[0] == h[0]

    b = np.asarray(h)

    assert b["value"][0] == h[0].value
    assert b["variance"][0] == h[0].variance

    h[0] = bh.accumulators.WeightedSum(value=3, variance=1)

    assert h[0].value == 3
    assert h[0].variance == 1

    assert a[0] == h[0]

    assert b["value"][0] == h[0].value
    assert b["variance"][0] == h[0].variance

    assert b[0]["value"] == a[0]["value"]
    assert b[0]["variance"] == a[0]["variance"]

    assert b["value"][0] == a["value"][0]
    assert b["variance"][0] == a["variance"][0]

    assert_array_equal(a.view().value, b.view()["value"])
    assert_array_equal(a.view().variance, b.view()["variance"])


def test_setting_profile():
    h = bh.Histogram(bh.axis.Regular(10, 0, 10), storage=bh.storage.Mean())

    h.fill([0.3, 0.3, 0.4, 1.2, 1.6], sample=[1, 2, 3, 4, 4])

    assert h[0] == bh.accumulators.Mean(count=3, value=2, variance=1)
    assert h[1] == bh.accumulators.Mean(count=2, value=4, variance=0)

    h[0] = bh.accumulators.Mean(count=12, value=11, variance=10)
    assert h[0] == bh.accumulators.Mean(count=12, value=11, variance=10)

    a = h.view()

    assert a[0] == h[0]

    b = np.asarray(h)

    assert b["value"][0] == h[0].value
    assert b["count"][0] == h[0].count
    assert b["sum_of_deltas_squared"][0] == h[0].sum_of_deltas_squared

    h[0] = bh.accumulators.Mean(count=6, value=3, variance=2)
    assert h[0].count == 6
    assert h[0].value == 3
    assert h[0].variance == 2

    assert a[0] == h[0]

    assert b["value"][0] == h[0].value
    assert b["count"][0] == h[0].count
    assert b["sum_of_deltas_squared"][0] == h[0].sum_of_deltas_squared

    assert b[0]["value"] == a[0]["value"]
    assert b[0]["count"] == a[0]["count"]
    assert b[0]["sum_of_deltas_squared"] == a[0]["sum_of_deltas_squared"]

    assert b[0]["value"] == a["value"][0]
    assert b[0]["count"] == a["count"][0]
    assert b[0]["sum_of_deltas_squared"] == a["sum_of_deltas_squared"][0]

    assert_array_equal(a.view().value, b.view()["value"])
    assert_array_equal(a.view().count, b.view()["count"])
    assert_array_equal(
        a.view().sum_of_deltas_squared, b.view()["sum_of_deltas_squared"]
    )


def test_setting_weighted_profile():
    h = bh.Histogram(bh.axis.Regular(10, 0, 10), storage=bh.storage.WeightedMean())

    h.fill([0.3, 0.3, 0.4, 1.2, 1.6], sample=[1, 2, 3, 4, 4], weight=[1, 1, 1, 1, 2])

    assert h[0] == bh.accumulators.WeightedMean(
        sum_of_weights=3, sum_of_weights_squared=3, value=2, variance=1
    )
    assert h[1] == bh.accumulators.WeightedMean(
        sum_of_weights=3, sum_of_weights_squared=5, value=4, variance=0
    )

    h[0] = bh.accumulators.WeightedMean(
        sum_of_weights=12, sum_of_weights_squared=15, value=11, variance=10
    )
    assert h[0] == bh.accumulators.WeightedMean(
        sum_of_weights=12, sum_of_weights_squared=15, value=11, variance=10
    )

    a = h.view()

    assert a[0] == h[0]

    b = np.asarray(h)

    assert b["value"][0] == h[0].value
    assert b["sum_of_weights"][0] == h[0].sum_of_weights
    assert b["sum_of_weights_squared"][0] == h[0].sum_of_weights_squared
    assert b["sum_of_weighted_deltas_squared"][0] == h[0].sum_of_weighted_deltas_squared

    h[0] = bh.accumulators.WeightedMean(
        sum_of_weights=6, sum_of_weights_squared=12, value=3, variance=2
    )

    assert a[0] == h[0]

    assert h[0].value == 3
    assert h[0].variance == 2
    assert h[0].sum_of_weights == 6
    assert h[0].sum_of_weights_squared == 12
    assert h[0].sum_of_weighted_deltas_squared == 8

    assert b["value"][0] == h[0].value
    assert b["sum_of_weights"][0] == h[0].sum_of_weights
    assert b["sum_of_weights_squared"][0] == h[0].sum_of_weights_squared
    assert b["sum_of_weighted_deltas_squared"][0] == h[0].sum_of_weighted_deltas_squared

    assert b[0]["value"] == a[0]["value"]
    assert b[0]["sum_of_weights"] == a[0]["sum_of_weights"]
    assert b[0]["sum_of_weights_squared"] == a[0]["sum_of_weights_squared"]
    assert (
        b[0]["sum_of_weighted_deltas_squared"] == a[0]["sum_of_weighted_deltas_squared"]
    )

    assert b[0]["value"] == a["value"][0]
    assert b[0]["sum_of_weights"] == a["sum_of_weights"][0]
    assert b[0]["sum_of_weights_squared"] == a["sum_of_weights_squared"][0]
    assert (
        b[0]["sum_of_weighted_deltas_squared"] == a["sum_of_weighted_deltas_squared"][0]
    )

    assert_array_equal(a.view().value, b.view()["value"])
    assert_array_equal(a.view().sum_of_weights, b.view()["sum_of_weights"])
    assert_array_equal(
        a.view().sum_of_weights_squared, b.view()["sum_of_weights_squared"]
    )
    assert_array_equal(
        a.view().sum_of_weighted_deltas_squared,
        b.view()["sum_of_weighted_deltas_squared"],
    )



from __future__ import absolute_import, division, print_function

from .kwargs import KWArgs

from .. import _core
from .view import _to_view
from .axis import Axis
from .axistuple import AxesTuple
from .sig_tools import inject_signature
from .storage import Double, Storage
from .utils import cast, register, set_family, MAIN_FAMILY, CPP_FAMILY, set_module

import warnings
import copy
import numpy as np

import os
import threading

_histograms = (
    _core.hist.any_double,
    _core.hist.any_int64,
    _core.hist.any_atomic_int64,
    _core.hist.any_unlimited,
    _core.hist.any_weight,
    _core.hist.any_mean,
    _core.hist.any_weighted_mean,
)


def _hist_or_val(other):
    return other._hist if hasattr(other, "_hist") else other


def _arg_shortcut(item):
    msg = "Developer shortcut: will be removed in a future version"
    if isinstance(item, tuple) and len(item) == 3:
        warnings.warn(msg, FutureWarning)
        return _core.axis.regular_uoflow(item[0], item[1], item[2], None)
    elif isinstance(item, tuple) and len(item) == 4:
        warnings.warn(msg, FutureWarning)
        return _core.axis.regular_uoflow(*item)
    elif isinstance(item, Axis):
        return item._ax
    else:
        raise TypeError("Only axes supported in histogram constructor")
        # TODO: Currently segfaults if we pass in a non-axis to the C++ code
        # Using the public interface above, this should never be possible.


def _expand_ellipsis(indexes, rank):
    indexes = list(indexes)
    number_ellipses = indexes.count(Ellipsis)
    if number_ellipses == 0:
        return indexes
    elif number_ellipses == 1:
        index = indexes.index(Ellipsis)
        additional = rank + 1 - len(indexes)
        if additional < 0:
            raise IndexError("too many indices for histogram")

        # Fill out the ellipsis with empty slices
        return indexes[:index] + [slice(None)] * additional + indexes[index + 1 :]

    else:
        raise IndexError("an index can only have a single ellipsis ('...')")


# We currently do not cast *to* a histogram, but this is consistent
# and could be used later.
@register(_histograms)
class BaseHistogram(object):
    @inject_signature("self, *axes, storage=Double()", locals={"Double": Double})
    def __init__(self, *axes, **kwargs):
        """
        Construct a new histogram.

        If you pass in a single argument, this will be treated as a
        histogram and this will convert the histogram to this type of
        histogram (DensityHistogram, Histogram, BoostHistogram).

        Parameters
        ----------
        *args : Axis
            Provide 1 or more axis instances.
        storage : Storage = bh.storage.Double()
            Select a storage to use in the histogram
        """

        # Allow construction from a raw histogram object (internal)
        if not kwargs and len(axes) == 1 and isinstance(axes[0], _histograms):
            self._hist = axes[0]
            return

        if not kwargs and len(axes) == 1 and isinstance(axes[0], BaseHistogram):
            self._hist = copy.copy(axes[0]._hist)
            return

        # Keyword only trick (change when Python2 is dropped)
        with KWArgs(kwargs) as k:
            storage = k.optional("storage", Double())

        # Check for missed parenthesis or incorrect types
        if not isinstance(storage, Storage):
            if issubclass(storage, Storage):
                raise KeyError(
                    "Passing in an initialized storage has been removed. Please add ()."
                )
            else:
                raise KeyError("Only storages allowed in storage argument")

        # Allow a tuple to represent a regular axis
        axes = [_arg_shortcut(arg) for arg in axes]

        if len(axes) > _core.hist._axes_limit:
            raise IndexError(
                "Too many axes, must be less than {}".format(_core.hist._axes_limit)
            )

        # Check all available histograms, and if the storage matches, return that one
        for h in _histograms:
            if isinstance(storage, h._storage_type):
                self._hist = h(axes, storage)
                return

        raise TypeError("Unsupported storage")

    def __array__(self):
        return _to_view(self._hist.view(False))

    def __add__(self, other):
        return self.__class__(self._hist.__add__(other._hist))

    def __iadd__(self, other):
        self._hist.__iadd__(other._hist)
        return self

    def __eq__(self, other):
        return self._hist == other._hist

    def __ne__(self, other):
        return self._hist != other._hist

    # If these fail, the underlying object throws the correct error
    def __mul__(self, other):
        return self.__class__(self._hist.__mul__(other))

    def __rmul__(self, other):
        return self * other

    def __imul__(self, other):
        self._hist.__imul__(_hist_or_val(other))
        return self

    def __truediv__(self, other):
        return self.__class__(self._hist.__truediv__(_hist_or_val(other)))

    def __div__(self, other):
        return self.__class__(self._hist.__div__(_hist_or_val(other)))

    def __itruediv__(self, other):
        self._hist.__itruediv__(_hist_or_val(other))
        return self

    def __idiv__(self, other):
        self._hist.__idiv__(_hist_or_val(other))
        return self

    def __copy__(self):
        other = self.__class__.__new__(self.__class__)
        other._hist = copy.copy(self._hist)
        return other

    @inject_signature("self, *args, weight=None, sample=None, threads=None")
    def fill(self, *args, **kwargs):
        """
        Insert data into the histogram.

        Parameters
        ----------
        *args : Union[Array[float], Array[int], Array[str], float, int, str]
            Provide one value or array per dimension.
        weight : List[Union[Array[float], Array[int], Array[str], float, int, str]]]
            Provide weights (only if the histogram storage supports it)
        sample : List[Union[Array[float], Array[int], Array[str], float, int, str]]]
            Provide samples (only if the histogram storage supports it)
        threads : Optional[int]
            Fill with threads. Defaults to None, which does not activate
            threaded filling.  Using 0 will automatically pick the number of
            available threads (usually two per core).
        """

        threads = kwargs.pop("threads", None)

        if threads is None or threads is 1:
            self._hist.fill(*args, **kwargs)
        else:
            if threads == 0:
                threads = os.cpu_count()

            print(self._hist._storage_type)
            if (
                self._hist._storage_type is _core.storage.mean
                or self._hist._storage_type is _core.storage.weighted_mean
            ):
                raise RuntimeError("Mean histograms do not support threaded filling")

            weight = kwargs.pop("weight", None)
            sample = kwargs.pop("sample", None)

            data = [np.array_split(a, threads) for a in args]

            if weight is None or np.isscalar(weight):
                weights = [weight] * threads
            else:
                weights = np.array_split(weight, threads)

            if sample is None or np.isscalar(sample):
                samples = [sample] * threads
            else:
                samples = np.array_split(sample, threads)

            if self._hist._storage_type is _core.storage.atomic_int64:

                def fun(weight, sample, *args):
                    self._hist.fill(*args, weight=weight, sample=sample)

            else:
                sum_lock = threading.Lock()

                def fun(weight, sample, *args):
                    local_hist = self._hist.__copy__()
                    local_hist.reset()
                    kw = {}
                    if weight is not None:
                        kw["weight"] = weight
                    if sample is not None:
                        kw["sample"] = sample
                    local_hist.fill(*args, **kw)
                    with sum_lock:
                        self._hist += local_hist

            thread_list = [
                threading.Thread(target=fun, args=arrays)
                for arrays in zip(weights, samples, *data)
            ]

            for thread in thread_list:
                thread.start()

            for thread in thread_list:
                thread.join()

        return self

    def __repr__(self):
        ret = "{self.__class__.__name__}(\n  ".format(self=self)
        ret += ",\n  ".join(repr(self._axis(i)) for i in range(self._hist.rank()))
        ret += ",\n  storage={0}".format(self._storage_type())
        ret += ")"
        outer = self._hist.sum(flow=True)
        if outer:
            inner = self._hist.sum(flow=False)
            ret += " # Sum: {0}".format(inner)
            if inner != outer:
                ret += " ({0} with flow)".format(outer)
        return ret

    def __str__(self):
        """
        A rendering of the histogram is made using ASCII or unicode characters (whatever is supported by the terminal). What exactly is displayed is still experimental. Do not rely on any particular rendering.
        """
        # TODO check the terminal width and adjust the presentation
        # only use for 1D, fall back to repr for ND
        if self._hist.rank() == 1:
            s = str(self._hist)
            # get rid of first line and last character
            s = s[s.index("\n") + 1 : -1]
        else:
            s = repr(self)
        return s

    def _axis(self, i):
        """
        Get N-th axis.
        """
        return cast(self, self._hist.axis(i), Axis)

    @property
    def _storage_type(self):
        return cast(self, self._hist._storage_type, Storage)

    def _reduce(self, *args):
        return self.__class__(self._hist.reduce(*args))


# C++ version of histogram
@set_family(CPP_FAMILY)
@set_module("boost_histogram.cpp")
class histogram(BaseHistogram):
    axis = BaseHistogram._axis

    def rank(self):
        """
        Number of axes (dimensions) of histogram.
        """
        return self._hist.rank()

    def size(self):
        """
        Total number of bins in the histogram (including underflow/overflow).
        """
        return self._hist.size()

    def at(self, *indexes):
        """
        Select a contents given indices. -1 is the underflow bin, N is the overflow bin.
        """
        return self._hist.at(*indexes)

    # Call uses fill since it supports strings,
    # runtime argument list, etc.
    @inject_signature("self, *args, weight=None, sample=None")
    def __call__(self, *args, **kwargs):
        args = (((a,) if isinstance(a, str) else a) for a in args)
        self._hist.fill(*args, **kwargs)
        return self

    def _reset(self):
        self._hist.reset()
        return self

    def _empty(self, flow=False):
        return self._hist.empty(flow)

    def _sum(self, flow=False):
        return self._hist.sum(flow)

    def _project(self, *args):
        return self.__class__(self._hist.project(*args))


@set_family(MAIN_FAMILY)
@set_module("boost_histogram")
class Histogram(BaseHistogram):
    @inject_signature("self, *axes, storage=Double()", locals={"Double": Double})
    def __init__(self, *args, **kwargs):
        super(Histogram, self).__init__(*args, **kwargs)

        # If this is a property, tab completion in IPython does not work
        self.axes = AxesTuple(self._axis(i) for i in range(self.rank))

    __init__.__doc__ = BaseHistogram.__init__.__doc__

    def __copy__(self):
        other = super(Histogram, self).__copy__()
        other.axes = AxesTuple(other._axis(i) for i in range(other.rank))
        return other

    def __deepcopy__(self, memo):
        other = self.__class__.__new__(self.__class__)
        other._hist = copy.deepcopy(self._hist, memo)
        other.axes = AxesTuple(other._axis(i) for i in range(other.rank))
        return other

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["axes"]  # Don't save the cashe
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.axes = AxesTuple(self._axis(i) for i in range(self.rank))

    def __repr__(self):
        newline = "\n  "
        sep = "," if len(self.axes) > 0 else ""
        ret = "{self.__class__.__name__}({newline}".format(
            self=self, newline=newline if len(self.axes) > 1 else ""
        )
        ret += ",{newline}".format(newline=newline).join(repr(ax) for ax in self.axes)
        ret += "{comma}{newline}storage={storage}".format(
            storage=self._storage_type(),
            newline=newline
            if len(self.axes) > 1
            else " "
            if len(self.axes) > 0
            else "",
            comma="," if len(self.axes) > 0 else "",
        )
        ret += ")"
        outer = self.sum(flow=True)
        if outer:
            inner = self.sum(flow=False)
            ret += " # Sum: {0}".format(inner)
            if inner != outer:
                ret += " ({0} with flow)".format(outer)
        return ret

    def _compute_commonindex(self, index):
        """
        Takes indices and returns two iterables; one is a tuple or dict of the
        original, Ellipsis expanded index, and the other returns index,
        operation value pairs.
        """
        # Shorten the computations with direct access to raw object
        hist = self._hist

        # Support dict access
        if hasattr(index, "items"):
            indexes = [slice(None)] * hist.rank()
            for k, v in index.items():
                indexes[k] = v

        # Normalize -> h[i] == h[i,]
        else:
            if not isinstance(index, tuple):
                index = (index,)
            # Now a list
            indexes = _expand_ellipsis(index, hist.rank())

        if len(indexes) != hist.rank():
            raise IndexError("Wrong number of indices for histogram")

        # Allow [bh.loc(...)] to work
        for i in range(len(indexes)):
            if callable(indexes[i]):
                indexes[i] = indexes[i](cast(self, hist.axis(i), Axis))
            elif hasattr(indexes[i], "flow"):
                if indexes[i].flow == 1:
                    indexes[i] = hist.axis(i).size
                elif indexes[i].flow == -1:
                    indexes[i] = -1
            elif isinstance(indexes[i], int):
                if abs(indexes[i]) >= hist.axis(i).size:
                    raise IndexError("histogram index is out of range")
                indexes[i] %= hist.axis(i).size

        return indexes

    def to_numpy(self, flow=False):
        """
        Convert to a Numpy style tuple of return arrays.

        Return
        ------
        contents : Array[Any]
            The bin contents
        *edges : Array[float]
            The edges for each dimension
        """
        return self._hist.to_numpy(flow)

    @inject_signature("self, *, deep=True")
    def copy(self, **kwargs):
        """
        Make a copy of the histogram. Defaults to making a
        deep copy (axis metadata copied); use deep=False
        to avoid making a copy of axis metadata.
        """

        # Future versions may add new options here
        with KWArgs(kwargs) as k:
            deep = k.optional("deep", True)

        if deep:
            return copy.deepcopy(self)
        else:
            return copy.copy(self)

    def view(self, flow=False):
        """
        Return a view into the data, optionally with overflow turned on.
        """
        return _to_view(self._hist.view(flow))

    def reset(self):
        """
        Reset bin counters to default values.
        """
        self._hist.reset()
        return self

    def empty(self, flow=False):
        """
        Check to see if the histogram has any non-default values.
        You can use flow=True to check flow bins too.
        """
        return self._hist.empty(flow)

    def sum(self, flow=False):
        """
        Compute the sum over the histogram bins (optionally including the flow bins).
        """
        return self._hist.sum(flow)

    @property
    def rank(self):
        """
        Number of axes (dimensions) of histogram.
        """
        return self._hist.rank()

    @property
    def size(self):
        """
        Total number of bins in the histogram (including underflow/overflow).
        """
        return self._hist.size()

    @property
    def shape(self):
        """
        Tuple of axis sizes (not including underflow/overflow).
        """
        return self.axes.size

    def __getitem__(self, index):

        indexes = self._compute_commonindex(index)

        # If this is (now) all integers, return the bin contents
        # But don't try *dict!
        if not hasattr(indexes, "items"):
            try:
                return self._hist.at(*indexes)
            except RuntimeError:
                pass

        integrations = set()
        slices = []
        zeroes_start = []
        zeroes_stop = []

        # We could use python's sum here, but for now, a private sum is used
        class ext_sum:
            projection = True

        # Compute needed slices and projections
        for i, ind in enumerate(indexes):
            if hasattr(ind, "__index__"):
                ind = slice(ind.__index__(), ind.__index__() + 1, ext_sum())

            elif not isinstance(ind, slice):
                raise IndexError(
                    "Must be a slice, an integer, or follow the locator protocol."
                )
            if ind != slice(None):
                merge = 1
                if ind.step is not None:
                    if hasattr(ind.step, "projection"):
                        if ind.step.projection:
                            integrations.add(i)
                            if ind.start is not None:  # TODO: Support callables too
                                zeroes_start.append(i)
                            if ind.stop is not None:
                                zeroes_stop.append(i)
                            if ind.stop is None and ind.start is None:
                                continue
                        elif hasattr(ind.step, "factor"):
                            merge = ind.step.factor
                        else:
                            raise IndexError("Invalid rebin, must have integer .factor")
                    else:
                        raise IndexError(
                            "The third argument to a slice must be rebin or projection"
                        )

                process_loc = (
                    lambda x, y: y
                    if x is None
                    else x(self._axis(i))
                    if callable(x)
                    else x
                )
                begin = process_loc(ind.start, 0)
                end = process_loc(ind.stop, len(self._axis(i)))

                slices.append(_core.algorithm.slice_and_rebin(i, begin, end, merge))

        reduced = self._reduce(*slices)

        if not integrations:
            return self.__class__(reduced)
        else:
            projections = [i for i in range(self.rank) if i not in integrations]

            # Replacement for crop missing in BH
            for i in zeroes_start:
                if self.axes[i].options.underflow:
                    reduced._hist._reset_row(i, -1)
            for i in zeroes_stop:
                if self.axes[i].options.overflow:
                    reduced._hist._reset_row(i, reduced.axes[i].size)

            return (
                self.__class__(reduced.project(*projections))
                if projections
                else reduced.sum(flow=True)
            )

    def __setitem__(self, index, value):
        """
        There are several supported possibilities:

            h[slice] = array # same size

        If an array is given to a compatible slice, it is set.

            h[a:] = array # One larger

        If an array is given that does not match, if it does match the
        with-overflow size, it fills that.

        PLANNED (not yet supported):

            h[a:] = h2

        If another histogram is given, that must either match with or without
        overflow, where the overflow bins must be overflow bins (that is,
        you cannot set a histogram's flow bins from another histogram that
        is 2 larger). Bin edges must be a close match, as well. If you don't
        want this level of type safety, just use ``h[...] = h2.view()``.
        """
        indexes = self._compute_commonindex(index)

        if isinstance(value, BaseHistogram):
            raise TypeError("Not supported yet")

        value = np.asarray(value)
        view = self.view(flow=True)

        # Disallow mismatched data types
        if len(value.dtype) != len(view.dtype):
            raise ValueError("Mismatched data types; matching types required")

        # Numpy does not broadcast partial slices, but we would need
        # to allow it (because we do allow broadcasting up dimensions)
        # Instead, we simply require matching dimensions.
        if value.ndim > 0 and value.ndim != sum(isinstance(i, slice) for i in indexes):
            raise ValueError(
                "Setting a {0}D histogram with a {1}D array must have a matching number of dimensions".format(
                    len(indexes), value.ndim
                )
            )

        # Here, value_n does not increment with n if this is not a slice
        value_n = 0
        for n, request in enumerate(indexes):
            has_underflow = self.axes[n].options.underflow
            has_overflow = self.axes[n].options.overflow

            if isinstance(request, slice):
                # Only consider underflow/overflow if the endpoints are not given
                use_underflow = has_underflow and request.start is None
                use_overflow = has_overflow and request.stop is None

                # Make the limits explicit since we may need to shift them
                start = 0 if request.start is None else request.start
                stop = len(self.axes[n]) if request.stop is None else request.stop
                request_len = stop - start

                # If set to a scalar, then treat it like broadcasting without flow bins
                if value.ndim == 0:
                    start = 0 + has_overflow
                    stop = len(self.axes[n]) + has_underflow

                # Normal setting
                elif request_len == value.shape[value_n]:
                    start += has_underflow
                    stop += has_underflow

                # Expanded setting
                elif request_len + use_underflow + use_overflow == value.shape[value_n]:
                    start += has_underflow and not use_underflow
                    stop += has_underflow + (has_overflow and use_overflow)

                # Single element broadcasting
                elif value.shape[value_n] == 1:
                    start += has_underflow
                    stop += has_underflow

                else:
                    msg = "Mismatched shapes in dimension {0}".format(n)
                    msg += ", {0} != {1}".format(value.shape[n], request_len)
                    if use_underflow or use_overflow:
                        msg += " or {0}".format(
                            request_len + use_underflow + use_overflow
                        )
                    raise ValueError(msg)
                indexes[n] = slice(start, stop, request.step)
                value_n += 1
            else:
                indexes[n] = request + has_underflow

        view[tuple(indexes)] = value

    def project(self, *args):
        """
        Project to a single axis or several axes on a multidiminsional histogram.
        Provided a list of axis numbers, this will produce the histogram over
        those axes only. Flow bins are used if available.
        """

        return self.__class__(self._hist.project(*args))



from __future__ import absolute_import, division, print_function

del absolute_import, division, print_function

__all__ = (
    "Regular",
    "Variable",
    "Integer",
    "IntCategory",
    "StrCategory",
    "Axis",
    "options",
    "transform",
)

from .._internal.axis import Axis, options
from .._internal.utils import register
from .._internal.axis import Regular, Variable, Integer, IntCategory, StrCategory
from . import transform



from __future__ import absolute_import, division, print_function

# Sadly, some tools (IPython) do not respect __all__
# as a list of public items in a module. So we need
# to delete / hide any extra items manually.
del absolute_import, division, print_function

__all__ = (
    "Histogram",
    "axis",
    "storage",
    "accumulators",
    "utils",
    "numpy",
    "loc",
    "rebin",
    "sum",
    "underflow",
    "overflow",
    "__version__",
)


try:
    from . import _core
except ImportError as err:
    if "_core" in err.msg and "boost_histogram" in err.msg:
        err.msg += "\nDid you forget to compile? Use CMake or Setuptools to build, see the readme"
    raise err


from ._internal.hist import Histogram
from . import axis, storage, accumulators, utils, numpy
from .tag import loc, rebin, sum, underflow, overflow

from .version import version as __version__



import boost_histogram as bh

import pytest


def test_subclass():
    NEW_FAMILY = object()

    @bh.utils.set_family(NEW_FAMILY)
    class MyHist(bh.Histogram):
        pass

    @bh.utils.set_family(NEW_FAMILY)
    class MyRegular(bh.axis.Regular):
        pass

    @bh.utils.set_family(NEW_FAMILY)
    class MyIntStorage(bh.storage.Int64):
        pass

    @bh.utils.set_family(NEW_FAMILY)
    class MyPowTransform(bh.axis.transform.Pow):
        pass

    h = MyHist(MyRegular(10, 0, 2, transform=MyPowTransform(2)), storage=MyIntStorage())

    assert type(h) == MyHist
    assert h._storage_type == MyIntStorage
    assert type(h.axes[0]) == MyRegular
    assert type(h.axes[0].transform) == MyPowTransform



# You can enter maximum compatibility mode by typing:
# import boost_histogram.cpp as bh
# However, this is a test so we will import both.
import boost_histogram as bh
import boost_histogram.cpp as bhc
from boost_histogram.cpp.algorithm import slice_mode
from numpy.testing import assert_array_equal
import numpy as np
import pytest


def test_usage_bh():
    h = bhc.histogram(bhc.axis.regular(10, 0, 1), bhc.axis.str_category(["one", "two"]))
    assert h.axis(0) == bhc.axis.regular(10, 0, 1)
    assert h.axis(1) == bhc.axis.str_category(["one", "two"])

    h(0, "one")

    assert h.at(0, 0) == 1.0


def test_convert_bh():
    h = bhc.histogram(
        bh.Histogram(bh.axis.Regular(10, 0, 1), bh.axis.StrCategory(["one", "two"]))
    )
    assert hasattr(h, "axis")
    assert not hasattr(h, "axes")

    h = bh.Histogram(h)

    # Current warning workaround. Enable when removed:
    # assert not hasattr(h, "axis")
    assert hasattr(h, "axes")


def test_str():
    # Mixing bh/bhc is fine; histogram correctly always returns matching axis,
    # axis returns matching transform, etc.
    h = bhc.histogram(bh.axis.Regular(4, 0, 4))
    assert repr(str(h)) == repr(
        """              +--------------------------------------------------------------+
[-inf,   0) 0 |                                                              |
[   0,   1) 0 |                                                              |
[   1,   2) 0 |                                                              |
[   2,   3) 0 |                                                              |
[   3,   4) 0 |                                                              |
[   4, inf) 0 |                                                              |
              +--------------------------------------------------------------+"""
    )


def test_repr():
    h = bhc.histogram(bh.axis.Regular(4, 0, 4))
    assert (
        repr(h)
        == """histogram(
  regular(4, 0, 4, metadata="None", options=underflow | overflow),
  storage=double())"""
    )


def test_axis_reprs():
    ax = bhc.axis.regular(4, 0, 4)
    assert repr(ax) == 'regular(4, 0, 4, metadata="None", options=underflow | overflow)'
    assert repr(type(ax)) == "<class 'boost_histogram.cpp.axis.regular'>"


def test_storage_repr():
    h = bhc.histogram(bhc.axis.regular(10, 0, 1))
    assert repr(h._storage_type()) == "double()"
    assert repr(h._storage_type) == "<class 'boost_histogram.cpp.storage.double'>"


def test_transform_repr():
    ax = bhc.axis.regular(8, 0, 3, transform=bhc.axis.transform.pow(2))
    assert repr(ax.transform()), "pow(2)"
    assert repr(
        type(ax.transform())
    ), "<class 'boost_histogram.cpp.axis.transform.pow'>"


def test_shrink():
    from boost_histogram.cpp.algorithm import reduce, shrink

    h = bhc.histogram(bhc.axis.regular(4, 1, 5))
    np.asarray(h)[:] = 1

    hs = reduce(h, shrink(0, 2, 3))
    assert hs.axis(0) == bhc.axis.regular(1, 2, 3)
    assert_array_equal(hs, [1])
    assert hs.at(-1) == 1
    assert hs.at(1) == 2

    hs2 = reduce(h, shrink(2, 3))
    assert hs == hs2


def test_crop():
    from boost_histogram.cpp.algorithm import reduce, crop

    h = bhc.histogram(bhc.axis.regular(4, 1, 5))
    np.asarray(h)[:] = 1

    hs = reduce(h, crop(0, 2, 3))
    assert hs.axis(0) == bhc.axis.regular(1, 2, 3)
    assert_array_equal(hs, [1])
    assert hs.at(-1) == 0
    assert hs.at(1) == 0

    hs2 = reduce(h, crop(2, 3))
    assert hs == hs2


@pytest.mark.parametrize("mode", (slice_mode.shrink, slice_mode.crop))
def test_slice(mode):
    from boost_histogram.cpp.algorithm import reduce, slice

    h = bhc.histogram(bhc.axis.regular(4, 1, 5))
    np.asarray(h)[:] = 1
    assert_array_equal(h, [1, 1, 1, 1])

    hs = reduce(h, slice(0, 1, 2, mode=mode))
    assert hs.axis(0) == bhc.axis.regular(1, 2, 3)
    assert_array_equal(hs, [1])
    assert hs.at(-1) == (1 if mode == slice_mode.shrink else 0)
    assert hs.at(1) == (2 if mode == slice_mode.shrink else 0)

    hs2 = reduce(h, slice(1, 2, mode=mode))
    assert hs == hs2


def test_rebin():
    from boost_histogram.cpp.algorithm import reduce, rebin

    h = bhc.histogram(bhc.axis.regular(4, 1, 5))
    np.asarray(h)[:] = 1
    assert_array_equal(h, [1, 1, 1, 1])

    hs = reduce(h, rebin(0, 4))
    assert hs.axis(0) == bhc.axis.regular(1, 1, 5)
    assert_array_equal(hs, [4])

    hs2 = reduce(h, rebin(4))
    assert hs == hs2


def test_shrink_and_rebin():
    from boost_histogram.cpp.algorithm import reduce, shrink_and_rebin

    h = bhc.histogram(bhc.axis.regular(5, 0, 5))
    np.asarray(h)[:] = 1
    hs = reduce(h, shrink_and_rebin(0, 1, 3, 2))
    assert hs.axis(0) == bhc.axis.regular(1, 1, 3)
    assert_array_equal(hs, [2])
    hs2 = reduce(h, shrink_and_rebin(1, 3, 2))
    assert hs == hs2


def test_crop_and_rebin():
    from boost_histogram.cpp.algorithm import reduce, crop_and_rebin

    h = bhc.histogram(bhc.axis.regular(5, 0, 5))
    np.asarray(h)[:] = 1
    hs = reduce(h, crop_and_rebin(0, 1, 3, 2))
    assert hs.axis(0) == bhc.axis.regular(1, 1, 3)
    assert_array_equal(hs, [2])
    hs2 = reduce(h, crop_and_rebin(1, 3, 2))
    assert hs == hs2


def test_slice_and_rebin():
    from boost_histogram.cpp.algorithm import reduce, slice_and_rebin

    h = bhc.histogram(bhc.axis.regular(5, 0, 5))
    np.asarray(h)[:] = 1
    hs = reduce(h, slice_and_rebin(0, 1, 3, 2))
    assert hs.axis(0) == bhc.axis.regular(1, 1, 3)
    assert_array_equal(hs, [2])
    hs2 = reduce(h, slice_and_rebin(1, 3, 2))
    assert hs == hs2



import pytest

import boost_histogram as bh

import numpy as np


def test_axes_all_at_once():
    h = bh.Histogram(
        bh.axis.Regular(10, 0, 10, metadata=2),
        bh.axis.Integer(0, 5, metadata="hi"),
        bh.axis.StrCategory(["HI", "HO"]),
    )

    assert h.axes.bin(1, 2, 0) == ((1.0, 2.0), 2, "HI")
    assert h.axes.value(0, 0, 0) == (0.0, 0, "HI")
    assert h.axes.index(2, 3, "HO") == (2, 3, 1)

    assert h.axes.size == (10, 5, 2)
    assert h.axes.extent == (12, 7, 3)
    assert h.axes.metadata == (2, "hi", None)

    h.axes.metadata = None, 3, "bye"

    assert h.axes.metadata == (None, 3, "bye")

    centers = h.axes.centers
    answers = np.ogrid[0.5:10, 0.5:5, 0.5:2]

    for i in range(3):
        np.testing.assert_allclose(centers[i], answers[i])
        np.testing.assert_allclose(h.axes[i].centers, answers[i].ravel())

    edges = h.axes.edges
    answers = np.ogrid[0:11, 0:6, 0:3]

    for i in range(3):
        np.testing.assert_allclose(edges[i], answers[i])
        np.testing.assert_allclose(h.axes[i].edges, answers[i].ravel())

    widths = h.axes.widths
    answers = np.ogrid[1:1:10j, 1:1:5j, 1:1:2j]

    for i in range(3):
        np.testing.assert_allclose(widths[i], answers[i])
        np.testing.assert_allclose(h.axes[i].widths, answers[i].ravel())



import boost_histogram as bh
import numpy as np

from numpy.testing import assert_array_equal

import pytest


def test_1D_set_bin():

    h = bh.Histogram(bh.axis.Regular(10, 0, 1))

    h[2] = 2
    assert h[2] == 2.0

    h[bh.underflow] = 3
    assert h[bh.underflow] == 3.0

    h[bh.overflow] = 4
    assert h[bh.overflow] == 4.0


def test_2d_set_bin():

    h = bh.Histogram(bh.axis.Regular(10, 0, 1), bh.axis.Regular(10, 0, 1))

    h[2, 4] = 2
    assert h[2, 4] == 2.0

    h[bh.underflow, 5] = 3
    assert h[bh.underflow, 5] == 3.0

    h[bh.overflow, bh.overflow] = 4
    assert h[bh.overflow, bh.overflow] == 4.0


def test_1d_set_array():
    h = bh.Histogram(bh.axis.Regular(10, 0, 1))

    h[...] = np.arange(10)
    assert_array_equal(h.view(), np.arange(10))

    h[...] = np.arange(12)
    assert_array_equal(h.view(flow=True), np.arange(12))

    with pytest.raises(ValueError):
        h[...] = np.arange(9)
    with pytest.raises(ValueError):
        h[...] = np.arange(11)
    with pytest.raises(ValueError):
        h[...] = np.arange(13)

    h[...] = 1
    assert_array_equal(h.view(), np.ones(10))


def test_2d_set_array():
    h = bh.Histogram(bh.axis.Regular(10, 0, 1), bh.axis.Regular(10, 0, 1))

    h[...] = np.arange(10).reshape(-1, 1)
    assert_array_equal(h.view()[:, 2], np.arange(10))

    h[...] = np.arange(12).reshape(-1, 1)
    assert_array_equal(h.view(flow=True)[:, 3], np.arange(12))

    with pytest.raises(ValueError):
        h[...] = np.arange(9).reshape(-1, 1)
    with pytest.raises(ValueError):
        h[...] = np.arange(11).reshape(-1, 1)
    with pytest.raises(ValueError):
        h[...] = np.arange(13).reshape(-1, 1)

    h[...] = 1
    assert_array_equal(h.view(), np.ones((10, 10)))


@pytest.mark.parametrize(
    "storage, default",
    (
        (bh.storage.Mean, bh.accumulators.Mean(1.0, 2.0, 3.0)),
        (bh.storage.WeightedMean, bh.accumulators.WeightedMean(1.0, 2.0, 3.0, 4.0)),
        (bh.storage.Weight, bh.accumulators.WeightedSum(1.0, 2)),
    ),
)
def test_set_special_dtype(storage, default):
    h = bh.Histogram(
        bh.axis.Regular(10, 0, 1), bh.axis.Regular(10, 0, 1), storage=storage()
    )

    arr = np.full((10, 1), default)
    h[...] = arr
    assert_array_equal(h.view()[:, 1:2], arr)

    arr = np.full((12, 1), default)
    h[...] = arr
    assert_array_equal(h.view(flow=True)[:, 2:3], arr)

    arr = np.full((10, 10), default)
    h[...] = arr
    assert_array_equal(h.view(), arr)

    arr = np.full((10, 12), default)
    h[...] = arr
    assert_array_equal(h.view(flow=True)[1:11, :], arr)

    arr = np.full((12, 10), default)
    h[...] = arr
    assert_array_equal(h.view(flow=True)[:, 1:11], arr)

    arr = np.full((12, 12), default)
    h[...] = arr
    assert_array_equal(h.view(flow=True), arr)

    with pytest.raises(ValueError):
        arr = np.full((9, 1), default)
        h[...] = arr
    with pytest.raises(ValueError):
        arr = np.full((11, 1), default)
        h[...] = arr
    with pytest.raises(ValueError):
        arr = np.full((13, 1), default)
        h[...] = arr

    with pytest.raises(ValueError):
        h[...] = 1

    with pytest.raises(ValueError):
        h[1, 1] = 1



from __future__ import absolute_import, division, print_function

del absolute_import, division, print_function

__all__ = ("Sum", "Mean", "WeightedSum", "WeightedMean")

from ._core.accumulators import Sum, Mean, WeightedSum, WeightedMean

for cls in (Sum, Mean, WeightedSum, WeightedMean):
    cls.__module__ = "boost_histogram.accumulators"
del cls

# Not supported by PyBind builtins
# Enable if wrapper added
# inject_signature("self, value")(Sum.fill)
# inject_signature("self, value, *, variance=None")(WeightedSum.fill)
# inject_signature("self, value, *, weight=None")(Mean.fill)
# inject_signature("self, value, *, weight=None")(WeightedMean.fill)



#!/usr/bin/env python3

import boost_histogram as bh
import pickle
from pathlib import Path

h1 = bh.Histogram(bh.axis.Regular(2, -1, 1))
h2 = h1.copy()

h1.fill(-0.5)
h2.fill(0.5)

# Arithmetic operators
h3 = h1 + h2
h4 = h3 * 2

print(f"{h4[0] = }, {h4[1] = }")

h4_saved = Path("h4_saved.pkl")

# Now save the histogram
with h4_saved.open("wb") as f:
    pickle.dump(h4, f, protocol=-1)

# And load
with h4_saved.open("rb") as f:
    h5 = pickle.load(f)

assert h4 == h5
print("Succeeded in pickling a histogram!")

# Delete the file to keep things tidy
h4_saved.unlink()



#!/usr/bin/env python3

import boost_histogram as bh

# Make 1-d histogram with 5 logarithmic bins from 1e0 to 1e5
h = bh.Histogram(
    bh.axis.Regular(5, 1e0, 1e5, metadata="x", transform=bh.axis.transform.log),
    storage=bh.storage.Weight(),
)

# Fill histogram with numbers
x = (2e0, 2e1, 2e2, 2e3, 2e4)

# Doing this several times so the variance is more interesting
h.fill(x, weight=2)
h.fill(x, weight=2)
h.fill(x, weight=2)
h.fill(x, weight=2)

# Iterate over bins and access bin counter
for idx, (lower, upper) in enumerate(h.axes[0]):
    val = h[idx]
    print(f"bin {idx} in [{lower:g}, {upper:g}): {val.value} +/- {val.variance**.5}")



#!/usr/bin/env python3

import numpy as np
import boost_histogram as bh
import matplotlib.pyplot as plt

# Make a 2D histogram
hist = bh.Histogram(bh.axis.Regular(50, -3, 3), bh.axis.Regular(50, -3, 3))

# Fill with Gaussian random values
hist.fill(np.random.normal(size=1_000_000), np.random.normal(size=1_000_000))

# Compute the areas of each bin
areas = np.prod(hist.axes.widths, axis=0)

# Compute the density
density = hist.view() / hist.sum() / areas

# Get the edges
X, Y = hist.axes.edges

# Make the plot
fig, ax = plt.subplots()
mesh = ax.pcolormesh(X.T, Y.T, density.T)
fig.colorbar(mesh)
plt.savefig("simple_density.png")



from __future__ import absolute_import, division, print_function

del absolute_import, division, print_function

__all__ = ("Slicer", "Locator", "at", "loc", "overflow", "underflow", "rebin", "sum")

import numpy as _np


class Slicer(object):
    """
    This is a simple class to make slicing inside dictionaries simpler.
    This is how it should be used:

        s = bh.tag.Slicer()

        h[{0: s[::bh.rebin(2)]}]   # rebin axis 0 by two

    """

    def __getitem__(self, item):
        return item


class Locator(object):
    __slots__ = ("offset",)

    def __init__(self, offset=0):
        if not isinstance(offset, int):
            raise ValueError("The offset must be an integer")

        self.offset = offset

    def __add__(self, offset):
        from copy import copy

        other = copy(self)
        other.offset += offset
        return other

    def __sub__(self, offset):
        from copy import copy

        other = copy(self)
        other.offset -= offset
        return other

    def _print_self_(self):
        return ""

    def __repr__(self):
        s = self.__class__.__name__
        s += self._print_self_()
        if self.offset != 0:
            s += " + " if self.offset > 0 else " - "
            s += str(abs(self.offset))
        return s


class loc(Locator):
    __slots__ = ("value",)

    def __init__(self, value, offset=0):
        super(loc, self).__init__(offset)
        self.value = value

    def _print_self_(self):
        return "({0})".format(self.value)

    def __call__(self, axis):
        return axis.index(self.value) + self.offset


class underflow(Locator):
    __slots__ = ()

    def __call__(self, axis):
        return -1 + self.offset


underflow = underflow()


class overflow(Locator):
    __slots__ = ()

    def __call__(self, axis):
        return len(axis) + self.offset


overflow = overflow()


class at(object):
    __slots__ = ("value",)

    def __init__(self, value):
        self.value = value

    def __call__(self, axis):
        if self.value < -2:
            raise IndexError("Index cannot be less than -1")

        return self.value


class rebin(object):
    __slots__ = ("factor",)
    projection = False

    def __init__(self, value):
        self.factor = value

    def __repr__(self):
        return "{self.__class__.__name__}({self.factor})".format(self=self)

    # TODO: Add __call__ to support UHI


class sum(object):
    __slots__ = ()
    projection = True

    # Supports UHI on general histograms, and acts nicely
    # if imported from boost_histogram directly
    def __new__(cls, *args, **kwargs):
        return _np.sum(*args, **kwargs)


# Compat only, will be removed after 0.6 release
class project(sum):
    pass



import pytest

import boost_histogram as bh
from boost_histogram._internal.sig_tools import make_signature_params

import inspect

pytestmark = pytest.mark.skipif(
    not hasattr(inspect, "Parameter"), reason="Python 3 only"
)


def test_simple_sigs():
    from inspect import Parameter

    assert make_signature_params("x, y") == [
        Parameter("x", Parameter.POSITIONAL_OR_KEYWORD),
        Parameter("y", Parameter.POSITIONAL_OR_KEYWORD),
    ]

    assert make_signature_params("x=True") == [
        Parameter("x", Parameter.POSITIONAL_OR_KEYWORD, default=True)
    ]
    assert make_signature_params("*, x=None") == [
        Parameter("x", Parameter.KEYWORD_ONLY, default=None)
    ]
    assert make_signature_params("y : 'wow'") == [
        Parameter("y", Parameter.POSITIONAL_OR_KEYWORD, annotation="wow")
    ]
    assert make_signature_params("y : None = 2") == [
        Parameter("y", Parameter.POSITIONAL_OR_KEYWORD, default=2, annotation=None)
    ]
    assert make_signature_params("*args, **kwargs") == [
        Parameter("args", Parameter.VAR_POSITIONAL),
        Parameter("kwargs", Parameter.VAR_KEYWORD),
    ]

    assert make_signature_params("*args, x={b'x':3}") == [
        Parameter("args", Parameter.VAR_POSITIONAL),
        Parameter("x", Parameter.KEYWORD_ONLY, default={b"x": 3}),
    ]


def test_fill_sig():
    from inspect import Parameter

    a, b, c, d, e = inspect.signature(bh.Histogram.fill).parameters.values()

    assert a == Parameter("self", Parameter.POSITIONAL_OR_KEYWORD)
    assert b == Parameter("args", Parameter.VAR_POSITIONAL)
    assert c == Parameter("weight", Parameter.KEYWORD_ONLY, default=None)
    assert d == Parameter("sample", Parameter.KEYWORD_ONLY, default=None)
    assert e == Parameter("threads", Parameter.KEYWORD_ONLY, default=None)



#!/usr/bin/env python3

import boost_histogram as bh

# make 1-d histogram with 5 logarithmic bins from 1e0 to 1e5
h = bh.Histogram(
    bh.axis.Regular(5, 1e0, 1e5, metadata="x", transform=bh.axis.transform.log),
    storage=bh.storage.Weight(),
)

# fill histogram with numbers
x = (2e0, 2e1, 2e2, 2e3, 2e4)
h.fill(x, weight=4)  # increment bin counter by 4

# iterate over bins and access bin counter
for idx, (lower, upper) in enumerate(h.axes[0]):
    val = h[idx]
    print(
        "bin {0} [{1:g}, {2:g}): {3} +/- {4}".format(
            idx, lower, upper, val.value, val.variance ** 0.5
        )
    )

# under- and overflow bin
lo, up = h.axes[0][bh.underflow]
print(
    "underflow [{0:g}, {1:g}): {2} +/- {3}".format(
        lo, up, h[bh.underflow].value, h[bh.underflow].variance ** 0.5
    )
)
lo, up = h.axes[0][bh.overflow]
print(
    "overflow  [{0:g}, {1:g}): {2} +/- {3}".format(
        lo, up, h[bh.overflow].value, h[bh.overflow].variance ** 0.5
    )
)

# prints:
# bin 0 x in [1.0, 10.0): 4.0 +/- 4.0
# bin 1 x in [10.0, 100.0): 4.0 +/- 4.0
# bin 2 x in [100.0, 1000.0): 4.0 +/- 4.0
# bin 3 x in [1000.0, 10000.0): 4.0 +/- 4.0
# bin 4 x in [10000.0, 100000.0): 4.0 +/- 4.0
# underflow [0.0, 1.0): 0.0 +/- 0.0
# overflow  [100000.0, inf): 0.0 +/- 0.0



import pytest
from pytest import approx

import boost_histogram as bh

import numpy as np
from numpy.testing import assert_array_equal, assert_allclose

import copy
import abc

# compatible with Python 2 *and* 3:
ABC = abc.ABCMeta("ABC", (object,), {"__slots__": ()})


@pytest.mark.parametrize(
    "axis,args,opt,kwargs",
    [
        (bh.axis.Regular, (1, 2, 3), "", {}),
        (bh.axis.Regular, (1, 2, 3), "u", {}),
        (bh.axis.Regular, (1, 2, 3), "o", {}),
        (bh.axis.Regular, (1, 2, 3), "uo", {}),
        (bh.axis.Regular, (1, 2, 3), "g", {}),
        (bh.axis.Regular, (1, 2, 3), "", {"circular": True}),
        (bh.axis.Regular, (1, 2, 3), "", {"transform": bh.axis.transform.log}),
        (bh.axis.Regular, (1, 2, 3), "", {"transform": bh.axis.transform.sqrt}),
        (bh.axis.Regular, (1, 2, 3), "", {"transform": bh.axis.transform.Pow(1)}),
        (bh.axis.Variable, ((1, 2, 3),), "", {}),
        (bh.axis.Variable, ((1, 2, 3),), "u", {}),
        (bh.axis.Variable, ((1, 2, 3),), "o", {}),
        (bh.axis.Variable, ((1, 2, 3),), "uo", {}),
        (bh.axis.Variable, ((1, 2, 3),), "g", {}),
        (bh.axis.Variable, ((1, 2, 3),), "", {"circular": True}),
        (bh.axis.Integer, (1, 2), "", {}),
        (bh.axis.Integer, (1, 2), "u", {}),
        (bh.axis.Integer, (1, 2), "o", {}),
        (bh.axis.Integer, (1, 2), "uo", {}),
        (bh.axis.Integer, (1, 2), "g", {}),
        (bh.axis.Integer, (1, 2), "", {"circular": True}),
        (bh.axis.IntCategory, ((1, 2, 3),), "", {}),
        (bh.axis.IntCategory, ((1, 2, 3),), "g", {}),
        (bh.axis.StrCategory, (tuple("ABC"),), "", {}),
        (bh.axis.StrCategory, (tuple("ABC"),), "g", {}),
    ],
)
def test_metadata(axis, args, opt, kwargs):
    for m in ("foo", 64, {"one": 1}):
        if "u" in opt:
            kwargs["underflow"] = False
        if "o" in opt:
            kwargs["overflow"] = False
        if "g" in opt:
            kwargs["growth"] = True
        kwargs["metadata"] = m

        assert axis(*args, **kwargs).metadata is m
        mcopy = copy.deepcopy(m)
        # assert axis(*args, m).metadata is not mcopy
        assert axis(*args, **kwargs).metadata == m
        assert axis(*args, **kwargs).metadata == mcopy
        assert axis(*args, **kwargs).metadata != "bar"
        assert axis(*args, **kwargs) == axis(*args, **kwargs)
        assert axis(*args, **kwargs) != axis(
            *args, metadata="bar"
        )  # TODO: Add original args


class Axis(ABC):
    @abc.abstractmethod
    def test_init(self):
        pass

    @abc.abstractmethod
    def test_equal(self):
        pass

    @abc.abstractmethod
    def test_len(self):
        pass

    @abc.abstractmethod
    def test_repr(self):
        pass

    @abc.abstractmethod
    def test_getitem(self):
        pass

    @abc.abstractmethod
    def test_iter(self):
        pass

    @abc.abstractmethod
    def test_index(self):
        pass

    # @abc.abstractmethod
    # def test_value(self):
    #     pass
    #
    @abc.abstractmethod
    def test_edges_centers_widths(self):
        pass


class TestRegular(Axis):
    def test_init(self):
        # Should not throw
        bh.axis.Regular(1, 1.0, 2.0)
        bh.axis.Regular(1, 1.0, 2.0, metadata="ra")
        bh.axis.Regular(1, 1.0, 2.0, underflow=False)
        bh.axis.Regular(1, 1.0, 2.0, underflow=False, overflow=False, metadata="ra")
        bh.axis.Regular(1, 1.0, 2.0, metadata=0)
        bh.axis.Regular(1, 1.0, 2.0, transform=bh.axis.transform.log)
        bh.axis.Regular(1, 1.0, 2.0, transform=bh.axis.transform.sqrt)
        bh.axis.Regular(1, 1.0, 2.0, transform=bh.axis.transform.Pow(1.5))

        with pytest.raises(TypeError):
            bh.axis.Regular()
        with pytest.raises(TypeError):
            bh.axis.Regular(overflow=False, underflow=False)
        with pytest.raises(TypeError):
            bh.axis.Regular(1)
        with pytest.raises(TypeError):
            bh.axis.Regular(1, 1.0)
        with pytest.raises(ValueError):
            bh.axis.Regular(0, 1.0, 2.0)
        with pytest.raises(TypeError):
            bh.axis.Regular("1", 1.0, 2.0)
        with pytest.raises(Exception):
            bh.axis.Regular(-1, 1.0, 2.0)

        with pytest.raises(ValueError):
            bh.axis.Regular(1, 1.0, 1.0)

        with pytest.raises(TypeError):
            bh.axis.Regular(1, 1.0, 2.0, bad_keyword="ra")
        with pytest.raises(AttributeError):
            bh.axis.Regular(1, 1.0, 2.0, transform=lambda x: 2)
        with pytest.raises(TypeError):
            bh.axis.Regular(1, 1.0, 2.0, transform=bh.axis.transform.Pow)
        # TODO: These errors could be better

        ax = bh.axis.Regular(1, 2, 3)
        assert isinstance(ax, bh.axis.Regular)
        assert ax.options == bh.axis.options(underflow=True, overflow=True)

        ax = bh.axis.Regular(1, 2, 3, overflow=False)
        assert isinstance(ax, bh.axis.Regular)
        assert ax.options == bh.axis.options(underflow=True)

        ax = bh.axis.Regular(1, 2, 3, underflow=False)
        assert isinstance(ax, bh.axis.Regular)
        assert ax.options == bh.axis.options(overflow=True)

        ax = bh.axis.Regular(1, 2, 3, underflow=False, overflow=False)
        assert isinstance(ax, bh.axis.Regular)
        assert ax.options == bh.axis.options()

        ax = bh.axis.Regular(1, 2, 3, growth=True)
        assert isinstance(ax, bh.axis.Regular)
        assert ax.options == bh.axis.options(underflow=True, overflow=True, growth=True)

    def test_equal(self):
        a = bh.axis.Regular(4, 1.0, 2.0)
        assert a == bh.axis.Regular(4, 1.0, 2.0)
        assert a != bh.axis.Regular(3, 1.0, 2.0)
        assert a != bh.axis.Regular(4, 1.1, 2.0)
        assert a != bh.axis.Regular(4, 1.0, 2.1)

        # metadata compare
        assert bh.axis.Regular(1, 2, 3, metadata=1) == bh.axis.Regular(
            1, 2, 3, metadata=1
        )
        assert bh.axis.Regular(1, 2, 3, metadata=1) != bh.axis.Regular(
            1, 2, 3, metadata="1"
        )
        assert bh.axis.Regular(1, 2, 3, metadata=1) != bh.axis.Regular(
            1, 2, 3, metadata=[1]
        )

    def test_len(self):
        a = bh.axis.Regular(4, 1.0, 2.0)
        assert len(a) == 4
        assert a.size == 4
        assert a.extent == 6

    def test_repr(self):
        ax = bh.axis.Regular(4, 1.1, 2.2)
        assert repr(ax) == "Regular(4, 1.1, 2.2)"

        ax = bh.axis.Regular(4, 1.1, 2.2, metadata="ra")
        assert repr(ax) == "Regular(4, 1.1, 2.2, metadata='ra')"

        ax = bh.axis.Regular(4, 1.1, 2.2, underflow=False)
        assert repr(ax) == "Regular(4, 1.1, 2.2, underflow=False)"

        ax = bh.axis.Regular(4, 1.1, 2.2, metadata="ra", overflow=False)
        assert repr(ax) == "Regular(4, 1.1, 2.2, overflow=False, metadata='ra')"

        ax = bh.axis.Regular(4, 1.1, 2.2, metadata="ra", circular=True)
        assert repr(ax) == "Regular(4, 1.1, 2.2, circular=True, metadata='ra')"

        ax = bh.axis.Regular(4, 1.1, 2.2, transform=bh.axis.transform.log)
        assert repr(ax) == "Regular(4, 1.1, 2.2, transform=log)"
        # TODO: Add caching so that an extracted functional transform actually works

        ax = bh.axis.Regular(3, 1.1, 2.2, transform=bh.axis.transform.sqrt)
        assert repr(ax) == "Regular(3, 1.1, 2.2, transform=sqrt)"

        ax = bh.axis.Regular(4, 1.1, 2.2, transform=bh.axis.transform.Pow(0.5))
        assert repr(ax) == "Regular(4, 1.1, 2.2, transform=pow(0.5))"

    def test_getitem(self):
        a = bh.axis.Regular(2, 1.0, 2.0)
        ref = [1.0, 1.5, 2.0]
        for i in range(2):
            assert_allclose(a.bin(i), ref[i : i + 2])
            assert_allclose(a[i], ref[i : i + 2])

        assert a[-1] == a[1]
        with pytest.raises(IndexError):
            a[2]

        assert a.bin(-1)[0] == -np.inf
        assert a.bin(2)[1] == np.inf

        assert_allclose(a[bh.underflow], a.bin(-1))
        assert_allclose(a[bh.overflow], a.bin(2))

        with pytest.raises(IndexError):
            a.bin(-2)
        with pytest.raises(IndexError):
            a.bin(3)

    def test_iter(self):
        a = bh.axis.Regular(2, 1.0, 2.0)
        ref = [(1.0, 1.5), (1.5, 2.0)]
        assert_allclose(a, ref)

    def test_index(self):
        a = bh.axis.Regular(4, 1.0, 2.0)

        assert a.index(-1) == -1
        assert a.index(0.99) == -1
        assert a.index(1.0) == 0
        assert a.index(1.249) == 0
        assert a.index(1.250) == 1
        assert a.index(1.499) == 1
        assert a.index(1.500) == 2
        assert a.index(1.749) == 2
        assert a.index(1.750) == 3
        assert a.index(1.999) == 3
        assert a.index(2.000) == 4
        assert a.index(20) == 4

    def test_reversed_index(self):
        a = bh.axis.Regular(4, 2.0, 1.0)

        assert a.index(-1) == 4
        assert a.index(0.99) == 4
        assert a.index(1.0) == 4
        assert a.index(1.249) == 3
        assert a.index(1.250) == 3
        assert a.index(1.499) == 2
        assert a.index(1.500) == 2
        assert a.index(1.749) == 1
        assert a.index(1.750) == 1
        assert a.index(1.999) == 0
        assert a.index(2.000) == 0
        assert a.index(20) == -1

    def test_sqrt_transform(self):
        a = bh.axis.Regular(10, 0, 10, transform=bh.axis.transform.sqrt)
        # Edges: 0. ,  0.1,  0.4,  0.9,  1.6,  2.5,  3.6,  4.9,  6.4,  8.1, 10.

        assert a.index(-100) == 10  # Always in overflow bin
        assert a.index(-1) == 10  # When transform is invalid
        assert a.index(0) == 0
        assert a.index(0.15) == 1
        assert a.index(0.5) == 2
        assert a.index(1) == 3
        assert a.index(1.7) == 4
        assert a.index(9) == 9
        assert a.index(11) == 10
        assert a.index(1000) == 10

        assert a.bin(0)[0] == approx(0)
        assert a.bin(1)[0] == approx(0.1)
        assert a.bin(1)[1] == approx(0.4)
        assert a.bin(2)[0] == approx(0.4)

    def test_log_transform(self):
        a = bh.axis.Regular(2, 1e0, 1e2, transform=bh.axis.transform.log)

        assert a.index(-1) == 2
        assert a.index(0.99) == -1
        assert a.index(1.0) == 0
        assert a.index(9.99) == 0
        assert a.index(10.0) == 1
        assert a.index(99.9) == 1
        assert a.index(100) == 2
        assert a.index(1000) == 2

        assert a.bin(0)[0] == approx(1e0)
        assert a.bin(1)[0] == approx(1e1)
        assert a.bin(1)[1] == approx(1e2)

    def test_pow_transform(self):
        a = bh.axis.Regular(2, 1.0, 9.0, transform=bh.axis.transform.Pow(0.5))

        assert a.index(-1) == 2
        assert a.index(0.99) == -1
        assert a.index(1.0) == 0
        assert a.index(3.99) == 0
        assert a.index(4.0) == 1
        assert a.index(8.99) == 1
        assert a.index(9) == 2
        assert a.index(1000) == 2

        assert a.bin(0)[0] == approx(1.0)
        assert a.bin(1)[0] == approx(4.0)
        assert a.bin(1)[1] == approx(9.0)

    def test_edges_centers_widths(self):
        a = bh.axis.Regular(2, 0, 1)
        assert_allclose(a.edges, [0, 0.5, 1])
        assert_allclose(a.centers, [0.25, 0.75])
        assert_allclose(a.widths, [0.5, 0.5])


class TestCircular(Axis):
    def test_init(self):
        # Should not throw
        bh.axis.Regular(1, 2, 3, circular=True)
        bh.axis.Regular(1, 2, 3, metadata="pa", circular=True)

        with pytest.raises(TypeError):
            bh.axis.Regular(1, 2, 3, "pa", circular=True)

        with pytest.raises(TypeError):
            bh.axis.Regular(circular=True)
        with pytest.raises(TypeError):
            bh.axis.Regular(1, circular=True)
        with pytest.raises(TypeError):
            bh.axis.Regular(1, -1, circular=True)
        with pytest.raises(Exception):
            bh.axis.Regular(-1, circular=True)
        with pytest.raises(TypeError):
            bh.axis.Regular(1, 1.0, metadata=1, circular=True)
        with pytest.raises(TypeError):
            bh.axis.Regular("1", circular=True)

    def test_equal(self):
        a = bh.axis.Regular(4, 0.0, 1.0, circular=True)
        assert a == bh.axis.Regular(4, 0, 1, circular=True)
        assert a != bh.axis.Regular(2, 0, 1, circular=True)
        assert isinstance(a, bh.axis.Regular)

    def test_len(self):
        assert len(bh.axis.Regular(4, 0.0, 1.0, circular=True)) == 4
        assert bh.axis.Regular(4, 0.0, 1.0, circular=True).size == 4
        assert bh.axis.Regular(4, 0.0, 1.0, circular=True).extent == 5

    def test_repr(self):
        ax = bh.axis.Regular(4, 1.1, 2.2, circular=True)
        assert repr(ax) == "Regular(4, 1.1, 2.2, circular=True)"

        ax = bh.axis.Regular(4, 1.1, 2.2, metadata="hi", circular=True)
        assert repr(ax) == "Regular(4, 1.1, 2.2, circular=True, metadata='hi')"

    def test_getitem(self):
        a = bh.axis.Regular(2, 1, 1 + np.pi * 2, circular=True)
        ref = [1.0, 1.0 + np.pi, 1.0 + 2.0 * np.pi]
        for i in range(2):
            assert_allclose(a.bin(i), ref[i : i + 2])
            assert_allclose(a[i], ref[i : i + 2])

        assert a[-1] == a[1]
        with pytest.raises(IndexError):
            a[2]

        with pytest.raises(IndexError):
            a[bh.underflow]

        assert_allclose(a[bh.overflow], a.bin(2))

        assert a.bin(2)[0] == approx(1 + 2 * np.pi)
        assert a.bin(2)[1] == approx(1 + 3 * np.pi)

        with pytest.raises(IndexError):
            a.bin(-1)  # no underflow
        with pytest.raises(IndexError):
            a.bin(3)

    def test_iter(self):
        a = bh.axis.Regular(2, 1, 2, circular=True)
        ref = [(1, 1.5), (1.5, 2)]
        assert_allclose(a, ref)

    def test_index(self):
        a = bh.axis.Regular(4, 1, 1 + np.pi * 2, circular=True)
        d = 0.5 * np.pi
        assert a.index(0.99 - 4 * d) == 3
        assert a.index(0.99 - 3 * d) == 0
        assert a.index(0.99 - 2 * d) == 1
        assert a.index(0.99 - d) == 2
        assert a.index(0.99) == 3
        assert a.index(1.0) == 0
        assert a.index(1.01) == 0
        assert a.index(0.99 + d) == 0
        assert a.index(1.0 + d) == 1
        assert a.index(1.0 + 2 * d) == 2
        assert a.index(1.0 + 3 * d) == 3
        assert a.index(1.0 + 4 * d) == 0
        assert a.index(1.0 + 5 * d) == 1

    def test_edges_centers_widths(self):
        a = bh.axis.Regular(2, 0, 1, circular=True)
        assert_allclose(a.edges, [0, 0.5, 1])
        assert_allclose(a.centers, [0.25, 0.75])
        assert_allclose(a.widths, [0.5, 0.5])


class TestVariable(Axis):
    def test_init(self):
        # should not raise
        bh.axis.Variable([0, 1])
        bh.axis.Variable((0, 1, 2, 3, 4))
        bh.axis.Variable([0, 1], metadata="va")
        with pytest.raises(TypeError):
            bh.axis.Variable()
        with pytest.raises(ValueError):
            bh.axis.Variable([1])
        with pytest.raises(TypeError):
            bh.axis.Variable(1)
        with pytest.raises(ValueError):
            bh.axis.Variable([1, -1])
        with pytest.raises(ValueError):
            bh.axis.Variable([1, 1])
        with pytest.raises(TypeError):
            bh.axis.Variable(["1", 2])
        with pytest.raises(TypeError):
            bh.axis.Variable([0.0, 1.0, 2.0], bad_keyword="ra")

        ax = bh.axis.Variable([1, 2, 3])
        assert isinstance(ax, bh.axis.Variable)
        assert ax.options == bh.axis.options(underflow=True, overflow=True)

        ax = bh.axis.Variable([1, 2, 3], overflow=False)
        assert isinstance(ax, bh.axis.Variable)
        assert ax.options == bh.axis.options(underflow=True)

        ax = bh.axis.Variable([1, 2, 3], underflow=False)
        assert isinstance(ax, bh.axis.Variable)
        assert ax.options == bh.axis.options(overflow=True)

        ax = bh.axis.Variable([1, 2, 3], underflow=False, overflow=False)
        assert isinstance(ax, bh.axis.Variable)
        assert ax.options == bh.axis.options()

    def test_equal(self):
        a = bh.axis.Variable([-0.1, 0.2, 0.3])
        assert a == bh.axis.Variable((-0.1, 0.2, 0.3))
        assert a != bh.axis.Variable([0, 0.2, 0.3])
        assert a != bh.axis.Variable([-0.1, 0.1, 0.3])
        assert a != bh.axis.Variable([-0.1, 0.1])

    def test_len(self):
        a = bh.axis.Variable([-0.1, 0.2, 0.3])
        assert len(a) == 2
        assert a.size == 2
        assert a.extent == 4

    def test_repr(self):
        a = bh.axis.Variable([-0.1, 0.2])
        assert repr(a) == "Variable([-0.1, 0.2])"

        a = bh.axis.Variable([-0.1, 0.2], metadata="hi")
        assert repr(a) == "Variable([-0.1, 0.2], metadata='hi')"

    def test_getitem(self):
        ref = [-0.1, 0.2, 0.3]
        a = bh.axis.Variable(ref)

        for i in range(2):
            assert_allclose(a.bin(i), ref[i : i + 2])
            assert_allclose(a[i], ref[i : i + 2])

        assert a[-1] == a[1]
        with pytest.raises(IndexError):
            a[2]

        assert_allclose(a[bh.underflow], a.bin(-1))
        assert_allclose(a[bh.overflow], a.bin(2))

        assert a.bin(-1)[0] == -np.inf
        assert a.bin(-1)[1] == ref[0]

        assert a.bin(2)[0] == ref[2]
        assert a.bin(2)[1] == np.inf

        with pytest.raises(IndexError):
            a.bin(-2)
        with pytest.raises(IndexError):
            a.bin(3)

    def test_iter(self):
        ref = [-0.1, 0.2, 0.3]
        a = bh.axis.Variable(ref)
        for i, bin in enumerate(a):
            assert_array_equal(bin, ref[i : i + 2])

    def test_index(self):
        a = bh.axis.Variable([-0.1, 0.2, 0.3])
        assert a.index(-10.0) == -1
        assert a.index(-0.11) == -1
        assert a.index(-0.1) == 0
        assert a.index(0.0) == 0
        assert a.index(0.19) == 0
        assert a.index(0.2) == 1
        assert a.index(0.21) == 1
        assert a.index(0.29) == 1
        assert a.index(0.3) == 2
        assert a.index(0.31) == 2
        assert a.index(10) == 2

    def test_edges_centers_widths(self):
        a = bh.axis.Variable([0, 1, 3])
        assert_allclose(a.edges, [0, 1, 3])
        assert_allclose(a.centers, [0.5, 2])
        assert_allclose(a.widths, [1, 2])


class TestInteger:
    def test_init(self):
        bh.axis.Integer(-1, 2)
        bh.axis.Integer(-1, 2, metadata="foo")
        bh.axis.Integer(-1, 2, underflow=False)
        bh.axis.Integer(-1, 2, underflow=False, overflow=False)
        bh.axis.Integer(-1, 2, growth=True)

        with pytest.raises(TypeError):
            bh.axis.Integer()
        with pytest.raises(TypeError):
            bh.axis.Integer(1)
        with pytest.raises(TypeError):
            bh.axis.Integer("1", 2)
        with pytest.raises(ValueError):
            bh.axis.Integer(2, -1)
        with pytest.raises(TypeError):
            bh.axis.Integer(-1, 2, "foo")
        with pytest.raises(TypeError):
            bh.axis.Integer(20, 30, 40)

        ax = bh.axis.Integer(1, 3)
        assert isinstance(ax, bh.axis.Integer)
        assert ax.options == bh.axis.options(underflow=True, overflow=True)

        ax = bh.axis.Integer(1, 3, overflow=False)
        assert isinstance(ax, bh.axis.Integer)
        assert ax.options == bh.axis.options(underflow=True)

        ax = bh.axis.Integer(1, 3, underflow=False)
        assert isinstance(ax, bh.axis.Integer)
        assert ax.options == bh.axis.options(overflow=True)

        ax = bh.axis.Integer(1, 3, underflow=False, overflow=False)
        assert isinstance(ax, bh.axis.Integer)
        assert ax.options == bh.axis.options()

        ax = bh.axis.Integer(1, 3, growth=True)
        assert isinstance(ax, bh.axis.Integer)
        assert ax.options == bh.axis.options(growth=True)

    def test_equal(self):
        assert bh.axis.Integer(-1, 2) == bh.axis.Integer(-1, 2)
        assert bh.axis.Integer(-1, 2) != bh.axis.Integer(-1, 2, metadata="Other")
        assert bh.axis.Integer(-1, 2, underflow=True) != bh.axis.Integer(
            -1, 2, underflow=False
        )

    @pytest.mark.parametrize("underflow", [0, 1])
    @pytest.mark.parametrize("overflow", [0, 1])
    def test_len(self, underflow, overflow):
        a = bh.axis.Integer(-1, 3, underflow=underflow, overflow=overflow)
        assert len(a) == 4
        assert a.size == 4
        assert a.extent == 4 + underflow + overflow

    def test_repr(self):
        a = bh.axis.Integer(-1, 1)
        assert repr(a) == "Integer(-1, 1)"

        a = bh.axis.Integer(-1, 1, metadata="hi")
        assert repr(a) == "Integer(-1, 1, metadata='hi')"

        a = bh.axis.Integer(-1, 1, underflow=False)
        assert repr(a) == "Integer(-1, 1, underflow=False)"

        a = bh.axis.Integer(-1, 1, overflow=False)
        assert repr(a) == "Integer(-1, 1, overflow=False)"

        a = bh.axis.Integer(-1, 1, growth=True)
        assert repr(a) == "Integer(-1, 1, growth=True)"

    def test_label(self):
        a = bh.axis.Integer(-1, 2, metadata="foo")
        assert a.metadata == "foo"
        a.metadata = "bar"
        assert a.metadata == "bar"

    def test_getitem(self):
        a = bh.axis.Integer(-1, 3)
        ref = [-1, 0, 1, 2]
        for i, r in enumerate(ref):
            assert a.bin(i) == r
            assert a[i] == r
        assert a.bin(-1) == -2
        assert a.bin(4) == 3

        assert_allclose(a[bh.underflow], a.bin(-1))
        assert_allclose(a[bh.overflow], a.bin(4))

    def test_iter(self):
        a = bh.axis.Integer(-1, 3)
        ref = (-1, 0, 1, 2)
        assert_array_equal(a, ref)

    def test_index(self):
        a = bh.axis.Integer(-1, 3)
        assert a.index(-3) == -1
        assert a.index(-2) == -1
        assert a.index(-1) == 0
        assert a.index(0) == 1
        assert a.index(1) == 2
        assert a.index(2) == 3
        assert a.index(3) == 4
        assert a.index(4) == 4

    def test_edges_centers_widths(self):
        a = bh.axis.Integer(1, 3)
        assert_allclose(a.edges, [1, 2, 3])
        assert_allclose(a.centers, [1.5, 2.5])
        assert_allclose(a.widths, [1, 1])


class TestCategory(Axis):
    def test_init(self):
        # should not raise
        bh.axis.IntCategory([1, 2])
        bh.axis.IntCategory((1, 2), metadata="foo")
        bh.axis.StrCategory(["A", "B"])
        bh.axis.StrCategory("AB")
        bh.axis.StrCategory("AB", metadata="foo")

        with pytest.raises(TypeError):
            bh.axis.IntCategory(["A", "B"])
        with pytest.raises(TypeError):
            bh.axis.StrCategory([1, 2])

        with pytest.raises(TypeError):
            bh.axis.IntCategory([1, 2], "foo")
        with pytest.raises(TypeError):
            bh.axis.StrCategory("AB", "foo")

        with pytest.raises(TypeError):
            bh.axis.StrCategory()
        with pytest.raises(TypeError):
            bh.axis.IntCategory()
        with pytest.raises(TypeError):
            bh.axis.StrCategory([1, "2"])
        with pytest.raises(TypeError):
            bh.axis.IntCategory([1, "2"])
        with pytest.raises(TypeError):
            bh.axis.IntCategory([1, 2, 3], underflow=True)

        ax = bh.axis.IntCategory([1, 2, 3])
        assert isinstance(ax, bh.axis.IntCategory)
        assert ax.options == bh.axis.options(overflow=True)

        ax = bh.axis.IntCategory([1, 2, 3], growth=True)
        assert isinstance(ax, bh.axis.IntCategory)
        assert ax.options == bh.axis.options(growth=True)

        ax = bh.axis.StrCategory(["1", "2", "3"])
        assert isinstance(ax, bh.axis.StrCategory)
        assert ax.options == bh.axis.options(overflow=True)

        ax = bh.axis.StrCategory(["1", "2", "3"], growth=True)
        assert isinstance(ax, bh.axis.StrCategory)
        assert ax.options == bh.axis.options(growth=True)

    def test_equal(self):
        assert bh.axis.IntCategory([1, 2, 3]) == bh.axis.IntCategory([1, 2, 3])
        assert bh.axis.IntCategory([1, 2, 3]) != bh.axis.IntCategory([1, 3, 2])
        assert bh.axis.StrCategory(["A", "B"]) == bh.axis.StrCategory("AB")
        assert bh.axis.StrCategory(["A", "B"]) != bh.axis.StrCategory("BA")

    @pytest.mark.parametrize("ref", ([1, 2, 3], "ABC"))
    @pytest.mark.parametrize("growth", (False, True))
    def test_len(self, ref, growth):
        Cat = bh.axis.StrCategory if isinstance(ref[0], str) else bh.axis.IntCategory
        a = Cat(ref, growth=growth)
        assert len(a) == 3
        assert a.size == 3
        assert a.extent == 3 if growth else 4

    def test_repr(self):
        ax = bh.axis.IntCategory([1, 2, 3])
        assert repr(ax) == "IntCategory([1, 2, 3])"

        ax = bh.axis.IntCategory([1, 2, 3], metadata="foo")
        assert repr(ax) == "IntCategory([1, 2, 3], metadata='foo')"

        ax = bh.axis.StrCategory("ABC", metadata="foo")
        # If unicode is the default (Python 3, generally)
        if type("") == type(u""):
            assert repr(ax) == "StrCategory(['A', 'B', 'C'], metadata='foo')"
        else:
            assert repr(ax) == "StrCategory([u'A', u'B', u'C'], metadata='foo')"

    @pytest.mark.parametrize("ref", ([1, 2, 3], "ABC"))
    @pytest.mark.parametrize("growth", (False, True))
    def test_getitem(self, ref, growth):
        Cat = bh.axis.StrCategory if isinstance(ref[0], str) else bh.axis.IntCategory

        a = Cat(ref, growth=growth)

        for i in range(3):
            assert a.bin(i) == ref[i]
            assert a[i] == ref[i]

        assert a[-1] == a[2]
        with pytest.raises(IndexError):
            a[3]

        with pytest.raises(IndexError):
            a.bin(-1)

        if growth:
            with pytest.raises(IndexError):
                a.bin(3)
        else:
            assert a.bin(3) == None

    @pytest.mark.parametrize("ref", ([1, 2, 3], ("A", "B", "C")))
    @pytest.mark.parametrize("growth", (False, True))
    def test_iter(self, ref, growth):
        Cat = bh.axis.StrCategory if isinstance(ref[0], str) else bh.axis.IntCategory
        a = Cat(ref, growth=growth)
        assert_array_equal(a, ref)

    @pytest.mark.parametrize("ref", ([1, 2, 3, 4], ("A", "B", "C", "D")))
    @pytest.mark.parametrize("growth", (False, True))
    def test_index(self, ref, growth):
        Cat = bh.axis.StrCategory if isinstance(ref[0], str) else bh.axis.IntCategory
        a = Cat(ref[:-1], growth=growth)
        for i, r in enumerate(ref):
            assert a.index(r) == i
        assert_array_equal(a.index(ref), [0, 1, 2, 3])
        assert_array_equal(a.index(np.reshape(ref, (2, 2))), [[0, 1], [2, 3]])

    @pytest.mark.parametrize("ref", ([1, 2, 3], ("A", "B", "C")))
    @pytest.mark.parametrize("growth", (False, True))
    def test_value(self, ref, growth):
        Cat = bh.axis.StrCategory if isinstance(ref[0], str) else bh.axis.IntCategory
        a = Cat(ref, growth=growth)
        for i, r in enumerate(ref):
            assert a.value(i) == r
        assert_array_equal(a.value(range(3)), ref)
        assert a.value(3) == None
        assert_array_equal(a.value((0, 3)), [ref[0], None])
        assert_array_equal(
            a.value(np.array((0, 1, 2, 3))), [ref[0], ref[1], ref[2], None]
        )
        # may be added in the future
        with pytest.raises(ValueError):
            a.value([[2], [2]])

    @pytest.mark.parametrize("ref", ([1, 2, 3], "ABC"))
    @pytest.mark.parametrize("growth", (False, True))
    def test_edges_centers_widths(self, ref, growth):
        Cat = bh.axis.StrCategory if isinstance(ref[0], str) else bh.axis.IntCategory
        a = Cat(ref, growth=growth)
        assert_allclose(a.edges, [0, 1, 2, 3])
        assert_allclose(a.centers, [0.5, 1.5, 2.5])
        assert_allclose(a.widths, [1, 1, 1])



from __future__ import absolute_import, division, print_function

del absolute_import, division, print_function

__all__ = ("sum", "mean", "weighted_sum", "weighted_mean")

from ..accumulators import (
    Sum as sum,
    Mean as mean,
    WeightedSum as weighted_sum,
    WeightedMean as weighted_mean,
)

# These will have the original module locations and original names.



import pytest
import numpy as np
from numpy.testing import assert_array_equal, assert_almost_equal

import ctypes

ftype = ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double)


def python_convert(x):
    return ftype(x)


try:
    # Python 2
    from cPickle import loads, dumps
except ImportError:
    from pickle import loads, dumps

import copy
import math

import boost_histogram as bh


def pickle_roundtrip_protocol_2(x):
    return loads(dumps(x, 2))


def pickle_roundtrip_protocol_highest(x):
    return loads(dumps(x, -1))


copy_fns = (
    pickle_roundtrip_protocol_2,
    pickle_roundtrip_protocol_highest,
    copy.copy,
    copy.deepcopy,
)


@pytest.mark.parametrize("copy_fn", copy_fns)
@pytest.mark.parametrize(
    "opts", ({}, {"growth": True}, {"underflow": True, "overflow": True})
)
def test_options(copy_fn, opts):
    orig = bh.axis.options(**opts)
    new = copy_fn(orig)
    assert new == orig


@pytest.mark.parametrize(
    "accum,args",
    (
        (bh.accumulators.Sum, (12,)),
        (bh.accumulators.WeightedSum, (1.5, 2.5)),
        (bh.accumulators.Mean, (5, 1.5, 2.5)),
        (bh.accumulators.WeightedMean, (1.5, 2.5, 3.5, 4.5)),
    ),
)
@pytest.mark.parametrize("copy_fn", copy_fns)
def test_accumulators(accum, args, copy_fn):
    orig = accum(*args)
    new = copy_fn(orig)
    assert new == orig


axes_creations = (
    (bh.axis.Regular, (4, 2, 4), {"underflow": False, "overflow": False}),
    (bh.axis.Regular, (4, 2, 4), {}),
    (bh.axis.Regular, (4, 2, 4), {"growth": True}),
    (bh.axis.Regular, (4, 2, 4), {"transform": bh.axis.transform.log}),
    (bh.axis.Regular, (4, 2, 4), {"transform": bh.axis.transform.sqrt}),
    (bh.axis.Regular, (4, 2, 4), {"transform": bh.axis.transform.Pow(0.5)}),
    (bh._core.axis.regular_numpy, (4, 2, 4), {}),
    (bh.axis.Regular, (4, 2, 4), {"circular": True}),
    (bh.axis.Variable, ([1, 2, 3, 4],), {}),
    (bh.axis.Variable, ([1, 2, 3, 4],), {"circular": True}),
    (bh.axis.Integer, (1, 4), {}),
    (bh.axis.Integer, (1, 4), {"circular": True}),
    (bh.axis.IntCategory, ([1, 2, 3],), {}),
    (bh.axis.IntCategory, ([1, 2, 3],), {"growth": True}),
    (bh.axis.StrCategory, (["1", "2", "3"],), {}),
    (bh.axis.StrCategory, (["1", "2", "3"],), {"growth": True}),
)


@pytest.mark.parametrize("axis,args,opts", axes_creations)
@pytest.mark.parametrize("copy_fn", copy_fns)
def test_axes(axis, args, opts, copy_fn):
    orig = axis(*args, metadata=None, **opts)
    new = copy_fn(orig)
    assert new == orig
    np.testing.assert_array_equal(new.centers, orig.centers)


@pytest.mark.parametrize("axis,args,opts", axes_creations)
@pytest.mark.parametrize("copy_fn", copy_fns)
def test_metadata_str(axis, args, opts, copy_fn):
    orig = axis(*args, metadata="foo", **opts)
    new = copy_fn(orig)
    assert new.metadata == orig.metadata
    new.metadata = orig.metadata
    assert new == orig
    np.testing.assert_array_equal(new.centers, orig.centers)


# Special test: Deepcopy should change metadata id, copy should not
@pytest.mark.parametrize("metadata", ({1: 2}, [1, 2, 3]))
def test_compare_copy_axis(metadata):
    orig = bh.axis.Regular(4, 0, 2, metadata=metadata)
    new = copy.copy(orig)
    dnew = copy.deepcopy(orig)

    assert orig.metadata is new.metadata
    assert orig.metadata == dnew.metadata
    assert orig.metadata is not dnew.metadata


# Special test: Deepcopy should change metadata id, copy should not
@pytest.mark.parametrize("metadata", ({1: 2}, [1, 2, 3]))
def test_compare_copy_hist(metadata):
    orig = bh.Histogram(bh.axis.Regular(4, 0, 2, metadata=metadata))
    new = copy.copy(orig)
    dnew = copy.deepcopy(orig)

    assert orig.axes[0].metadata is new.axes[0].metadata
    assert orig.axes[0].metadata == dnew.axes[0].metadata
    assert orig.axes[0].metadata is not dnew.axes[0].metadata


@pytest.mark.parametrize("axis,args,opts", axes_creations)
@pytest.mark.parametrize("copy_fn", copy_fns)
def test_metadata_any(axis, args, opts, copy_fn):
    orig = axis(*args, metadata=(1, 2, 3), **opts)
    new = copy_fn(orig)
    assert new.metadata == orig.metadata
    new.metadata = orig.metadata
    assert new == orig


@pytest.mark.benchmark(group="histogram-pickling")
@pytest.mark.parametrize("copy_fn", copy_fns)
@pytest.mark.parametrize(
    "storage, extra",
    (
        (bh.storage.AtomicInt64, {}),
        (bh.storage.Int64, {}),
        (bh.storage.Unlimited, {}),
        (bh.storage.Unlimited, {"weight"}),
        (bh.storage.Double, {"weight"}),
        (bh.storage.Weight, {"weight"}),
        (bh.storage.Mean, {"sample"}),
        (bh.storage.WeightedMean, {"weight", "sample"}),
    ),
)
def test_storage(benchmark, copy_fn, storage, extra):
    n = 1000
    hist = bh.Histogram(bh.axis.Integer(0, n), storage=storage())
    x = np.arange(2 * (n + 2)) % (n + 2) - 1
    if extra == {}:
        hist.fill(x)
    elif extra == {"weight"}:
        hist.fill(x, weight=np.arange(2 * n + 4) + 1)
    elif extra == {"sample"}:
        hist.fill(x, sample=np.arange(2 * n + 4) + 1)
    else:
        hist.fill(x, weight=np.arange(2 * n + 4) + 1, sample=np.arange(2 * n + 4) + 1)

    new = benchmark(copy_fn, hist)
    assert_array_equal(hist.view(True), new.view(True))
    assert new == hist


@pytest.mark.parametrize("copy_fn", copy_fns)
def test_histogram_regular(copy_fn):
    hist = bh.Histogram(bh.axis.Regular(4, 1, 2), bh.axis.Regular(8, 3, 6))

    new = copy_fn(hist)
    assert hist == new


@pytest.mark.parametrize("copy_fn", copy_fns)
def test_histogram_fancy(copy_fn):
    hist = bh.Histogram(bh.axis.Regular(4, 1, 2), bh.axis.Integer(0, 6))

    new = copy_fn(hist)
    assert hist == new


@pytest.mark.parametrize("copy_fn", copy_fns)
@pytest.mark.parametrize("metadata", ("This", (1, 2, 3)))
def test_histogram_metadata(copy_fn, metadata):

    hist = bh.Histogram(bh.axis.Regular(4, 1, 2, metadata=metadata))
    new = copy_fn(hist)
    assert hist == new


@pytest.mark.parametrize("copy_fn", copy_fns)
def test_numpy_edge(copy_fn):
    ax1 = bh._core.axis.regular_numpy(10, 0, 1, None)
    ax2 = copy_fn(ax1)

    # stop defaults to 0, so this fails if the copy fails
    assert ax1 == ax2
    assert ax1.index(1) == ax2.index(1)
    assert ax2.index(1) == 9


@pytest.mark.parametrize("mod", (np, math))
@pytest.mark.parametrize("copy_fn", copy_fns)
def test_pickle_transforms(mod, copy_fn):
    ax1 = bh.axis.Regular(
        100,
        1,
        100,
        transform=bh.axis.transform.Function(mod.log, mod.exp, convert=python_convert),
    )
    ax2 = copy_fn(ax1)
    ax3 = bh.axis.Regular(100, 1, 100, transform=bh.axis.transform.log)

    assert ax1 == ax2
    assert_array_equal(ax1.centers, ax2.centers)
    assert_almost_equal(ax2.centers, ax3.centers, decimal=10)


@pytest.mark.parametrize("copy_fn", copy_fns)
def test_hist_axes_reference(copy_fn):
    h = bh.Histogram(bh.axis.Regular(10, 0, 1, metadata=1))
    h.axes[0].metadata = 2

    # Note: copy does not copy the metadata, but we are *replacing*
    # the metadata here, so it should be distinct even after a shallow copy.
    h2 = copy_fn(h)

    assert h2._hist is not h._hist
    assert h2.axes[0] is not h.axes[0]

    assert h2.axes[0].metadata == 2

    h.axes[0].metadata = 3
    assert h2._axis(0).metadata == 2
    assert h2.axes[0].metadata == 2


@pytest.mark.parametrize("copy_fn", copy_fns)
def test_axis_wrapped(copy_fn):
    ax = bh.axis.Regular(10, 0, 2)
    ax2 = copy_fn(ax)

    assert ax._ax is not ax2._ax


@pytest.mark.parametrize("copy_fn", copy_fns)
def test_trans_wrapped(copy_fn):
    tr = bh.axis.transform.Pow(2)
    tr2 = copy_fn(tr)

    assert tr._this is not tr2._this



from __future__ import absolute_import, division, print_function

del absolute_import, division, print_function

from .._core import axis as ca

from .utils import register, set_family, CPP_FAMILY, MAIN_FAMILY, set_module
from .sig_tools import inject_signature
from .kwargs import KWArgs

import copy


@set_module("boost_histogram.axis.transform")
class AxisTransform(object):
    __slots__ = ("_this",)

    def __copy__(self):
        other = self.__class__.__new__(self.__class__)
        other._this = copy.copy(self._this)
        return other

    @classmethod
    def _convert_cpp(cls, this):
        self = cls.__new__(cls)
        self._this = this
        return self

    def __repr__(self):
        if hasattr(self, "_this"):
            return repr(self._this)
        else:
            return self.__class__.__name__ + "() # Missing _this, broken class"

    def _produce(self, bins, start, stop, metadata):
        return self.__class__._type(bins, start, stop, metadata)

    def __init__(self):
        "Create a new transform instance"
        cpp_class, = self._types
        self._this = cpp_class()

    def forward(self, value):
        "Compute the forward transform"
        return self._this.forward(value)

    def inverse(self, value):
        "Compute the inverse transform"
        return self._this.inverse(value)


core = "__init__ forward inverse".split()


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.axis.transform")
@register({ca.transform.pow})
class Pow(AxisTransform):
    __slots__ = ()
    _type = ca.regular_pow

    def __init__(self, power):
        "Create a new transform instance"
        cpp_class, = self._types
        self._this = cpp_class(power)

    @property
    def power(self):
        "The power of the transform"
        return self._this.power

    # This one does need to be a normal method
    def _produce(self, bins, start, stop, metadata):
        return self.__class__._type(bins, start, stop, self.power, metadata)


@set_family(MAIN_FAMILY)
@set_module("boost_histogram.axis.transform")
@register({ca.transform.func_transform})
class Function(AxisTransform):
    __slots__ = ()
    _type = ca.regular_trans

    @inject_signature("self, forward, inverse, *, convert=None, name=''")
    def __init__(self, forward, inverse, **kwargs):
        """
        Create a functional transform from a ctypes double(double) function
        pointer or any object that provides such an interface through a
        ``.ctypes`` attribute (such as numba.cfunc). A pure python function *can*
        be adapted to a ctypes pointer, but please use a Variable axis instead or
        use something like numba to produce a compiled function pointer. You can
        manually specify the repr name with ``name=``.

        Example of Numba use:
        ---------------------

            @numba.cfunc(numba.float64(numba.float64,))
            def exp(x):
                return math.exp(x)

            @numba.cfunc(numba.float64(numba.float64,))
            def log(x):
                return math.log(x)

        Example of slow CTypes use:
        ---------------------------

            ftype = ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double)
            log = ftype(math.log)
            exp = ftype(math.exp)


        Now you can supply these functions, and you will get a high performance
        transformation axis.

        You can also supply an optional conversion function; this will take the input
        forward and inverse and call them before producing a transform. This enables
        pickling, as well, since ctypes pointers are not picklable. A few common
        utilities have been supplied:

        * ``convert.numba``: Compile using numba (required)
        * ``convert.python``: Just call the Python function (15-90x slower than compiled)

        See also
        --------

        * ``Numbify(forward, inverse, *, name='')``: Uses convert=convert.numba
        * ``PythonFunction(forward, inverse, *, name='')``: Uses convert=convert.python

        """

        with KWArgs(kwargs) as k:
            convert = k.optional("convert")
            name = k.optional("name", "")

        cpp_class, = self._types
        self._this = cpp_class(forward, inverse, convert, name)

    # This one does need to be a normal method
    def _produce(self, bins, start, stop, metadata):
        return self.__class__._type(bins, start, stop, self._this, metadata)


def _internal_conversion(value):
    return getattr(ca.transform, value)


### CPP FAMILY ###

# Not currently included, uses normal family for now.
